

from traceback import print_exc
from contextlib import contextmanager
from kivy.graphics import Fbo, Color, Rectangle
from kivy.core.window import Window
from kivy.clock import Clock
from kivy.utils import platform
from kivy.app import App
from kivy.properties import BooleanProperty
from plyer import filechooser

desktopOSs=[
	"win",
	"linux",
	"macosx"
]

class WidgetReady(object):
	fullscreen = BooleanProperty(False)
	ready = BooleanProperty(False)
	_fullscreen = False

	@contextmanager
	def fboContext(self):
		self._fbo = Fbo(size=self.size)
		with self._fbo:
			self._background_color = Color(0,0,0,1)
			self._background_rect = Rectangle(size=self.size)

		try:
			yield self._fbo
		except Exception as e:
			print_exc()
			print('Exeception=',e)
			
		with self.canvas:
			self._fbo_rect = Rectangle(size=self.size,
								texture=self._fbo.texture)

	def on_ready(self, *args):
		pass

	def set_ready(self, *args):
		self.ready = True

	def reready(self):
		self.ready = False
		Clock.schedule_once(self.set_ready, 0.1)

	def use_keyboard(self):
		self.my_kb = Window.request_keyboard(None, self)
		if not self.my_kb:
			print('my_kb is None........')
			return 
		self.my_kb.bind(on_key_down=self._on_keyboard_down)
		if self.my_kb.widget:
			self.my_kb.set_mode_free()

	def key_handle(self,keyinfo):
		pass

	def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
		print('The key', keycode, 'have been pressed')
		print(' - text is %r' % text)
		print(' - modifiers are %r' % modifiers)
		keyinfo = {
			"keyname":keycode[1],
			"modifiers":modifiers
		}
		self.key_handle(keyinfo)
		return True

	def on_fullscreen(self, instance, value):
		window = self.get_parent_window()
		if not window:
			Logger.warning('VideoPlayer: Cannot switch to fullscreen, '
						   'window not found.')
			return
		if not self.parent:
			Logger.warning('VideoPlayer: Cannot switch to fullscreen, '
						   'no parent.')
			return

		app = App.get_running_app()
		if value:
			Window.fullscreen = True
			app.fs_widget = self
			self._fullscreen_state = state = {
				'parent': self.parent,
				'pos': self.pos,
				'size': self.size,
				'pos_hint': self.pos_hint,
				'size_hint': self.size_hint,
				'window_children': window.children[:]}

			# if platform in desktopOSs:
			# 	Window.maximize()
			# remove all window children
			for child in window.children[:]:
				window.remove_widget(child)

			if state['parent'] is not window:
				state['parent'].remove_widget(self)
			window.add_widget(self)

			self.pos = (0, 0)
			self.pos_hint = {}
			self.size_hint = (1, 1)
		else:
			app.fs_widget = None
			Window.fullscreen = False
			#if platform in desktopOSs:
			#	Window.restore()
			state = self._fullscreen_state
			window.remove_widget(self)
			for c in state['window_children']:
				if c in Window.children:
					Window.remove_widget(c)
			for child in state['window_children']:
				window.add_widget(child)
			self.pos_hint = state['pos_hint']
			self.size_hint = state['size_hint']
			self.pos = state['pos']
			self.size = state['size']
			if state['parent'] is not window:
				state['parent'].add_widget(self)

	def file_open_for_read(self, *args, **kw):
		method_name = kw.get('on_selection')
		if method_name is None:
			return
		f = getattr(self, method_name)
		if f is None:
			return 
		kw['on_selection'] = f
		filechooser.open_file(**kw)

	def file_open_for_write(self, *args, **kw):
		method_name = kw.get('on_selection')
		if method_name is None:
			return
		f = getattr(self, method_name)
		if f is None:
			return 
		kw['on_selection'] = f
		filechooser.save_file(**kw)
		




from ..utils import *
from ..dateinput import DateInput
from .factory import UiFactory, StringView, get_value

def build_input_date_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = DateInput(**kw)
	if v:
		w.setValue(v)
	return w

UiFactory.register('date', build_input_date_widget, StringView.builder)




try:
	import ujson as json
except:
	import json
from kivy.factory import Factory
from .factory import UiFactory
from appPublic.myTE import MyTemplateEngine

def view_build_viewer(desc, rec=None):
	viewer_desc = desc.get('viewer')
	if not viewer_desc:
		return None
	if isinstance(viewer_desc, dict):
		viewer_desc = json.dumps(viewer_desc)

	mytmpl = MyTemplateEngine(['.'])
	v_desc = mytmpl.renders(viewer_desc, rec)
	vdesc = json.loads(v_desc)
	return Factory.Blocks().widgetBuild(vdesc)

UiFactory.register('viewer', None, view_build_viewer)



import tempfile
from kivyblocks.baseWidget import HTTPDataHandler
from kivyblocks.utils import blockImage, absurl

from kivy.uix.image import Image

class HostImage(Image):
	def __init__(self, target,**kwargs):
		self.options = kwargs
		self.target = target
		kwargs['source'] = blockImage('running.gif')
		url = kwargs.get('url')
		del kwargs['url']
		super().__init__(**kwargs)
		self.downloadImage(url)

	def downloadImage(self,url):
		realurl = absurl(url,self.target.parenturl)
		loader = HTTPDataHandler(url,stream=True)
		loader.bind(on_success=self.createTmpfile)
		loader.bind(on_failed=self.showBadImage)
		loader.handle()
	
	def showBadImage(self,o,e):
		self.source = blockImage('broken.png')
		
	def createTmpfile(self,o,resp):
		fn = tempfile.mkstemp()[1]
		print('************fn=%s',fn)
		with open(fn, 'wb') as f:
			for chunk in resp.iter_content(chunk_size=8192): 
				if chunk: # filter out keep-alive new chunks
					f.write(chunk)
					# f.flush()
		self.source = fn




from pygments.lexers.data import JsonLexer
from kivy.uix.codeinput import CodeInput

class JsonCodeInput(CodeInput):
	def __init__(self,**kw):
		super(JsonCodeInput,self).__init__(lexer = JsonLexer())



class PlayListBase:
	def __init__(self, player, url, params, curpos):
		self.dataloader = RelatedLoader
	def first(self):
		pass

	def next(self):
		pass

	def previous(self):
		pass

	def last(self):
		pass
	
	def getSource(self):
		pass





''' Kivy Widget that accepts data and displays qrcode
'''

from threading import Thread
from functools import partial

from kivy.factory import Factory
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.image import Image
from kivy.graphics.texture import Texture
from kivy.properties import StringProperty
from kivy.properties import ObjectProperty, StringProperty, ListProperty,\
	BooleanProperty, NumericProperty
from kivy.lang import Builder
from kivy.clock import Clock

from .baseWidget import VBox
import qrcode

class QRCodeWidget(VBox):
	data = StringProperty(None, allow_none=True)
	''' Data using which the qrcode is generated.

	:data:`data` is a :class:`~kivy.properties.StringProperty`, defaulting to
	`None`.
	'''

	loading_image = StringProperty('data/images/image-loading.gif')
	'''Intermediate image to be displayed while the widget ios being loaded.
	
	:data:`loading_image` is a :class:`~kivy.properties.StringProperty`,
	defaulting to `'data/images/image-loading.gif'`.
	'''

	version = NumericProperty(40)
	box_size = NumericProperty(20)
	err_level = StringProperty('L')
	border = NumericProperty(1)
	error_correct_values = {
		'L':qrcode.constants.ERROR_CORRECT_L,
		'M':qrcode.constants.ERROR_CORRECT_M,
		'Q':qrcode.constants.ERROR_CORRECT_Q,
		'H':qrcode.constants.ERROR_CORRECT_H
	}
	def __init__(self, **kwargs):
		self.qrimage = Image(allow_stretch=True, keep_ratio=True)
		self.addr = None
		super(QRCodeWidget, self).__init__(**kwargs)
		self.background_color = [1,1,1,1]
		self.qr = None
		self._qrtexture = None
		self.add_widget(self.qrimage)
	
	def on_size(self,o,s):
		self.qrimage.width = self.width
		self.qrimage.height = self.height
		self.on_data(None,self.data)

	def on_data(self, instance, value):
		if not (self.canvas or value):
			return
		img = self.qrimage
		img.anim_delay = .25
		img.source = self.loading_image
		self.generate_qr(value)

	def generate_qr(self, value):
		self.set_addr(value)
		self.update_qr()

	def set_addr(self, addr):
		if self.addr == addr:
			return
		MinSize = 500 #210 if len(addr) < 128 else 500
		self.setMinimumSize((MinSize, MinSize))
		self.addr = addr
		self.qr = None

	def update_qr(self):
		if not self.addr and self.qr:
			return
		QRCode = qrcode.QRCode
		L = qrcode.constants.ERROR_CORRECT_L
		addr = self.addr
		print('self.box_size=', self.box_size)
		errv = self.error_correct_values.get(self.err_level)
		if not errv:
			errv=self.error_corrent_values.get('L')

		try:
			self.qr = qr = QRCode(
				version=self.version,
				error_correction=errv,
				box_size=self.box_size,
				border=self.border,
				)
			qr.add_data(addr)
			qr.make(fit=True)
			self.qr = qr
			self.update_texture()
		except Exception as e:
			print('eeeee',e)
			self.qr=None

	def setMinimumSize(self, size):
		# currently unused, do we need this?
		self._texture_size = size

	def _create_texture(self, k):
		self._qrtexture = texture = Texture.create(size=(k,k), colorfmt='rgb')
		# don't interpolate texture
		texture.min_filter = 'nearest'
		texture.mag_filter = 'nearest'

	def update_texture(self):
		if not self.addr:
			return

		matrix = self.qr.get_matrix()
		k = len(matrix)
		self._create_texture(k)
		
		
		cr, cg, cb = 255, 255, 255
		###used bytearray for python 3.5 eliminates need for btext
		buff = bytearray()
		for r in range(k):
			for c in range(k):
				buff.extend([0, 0, 0] if matrix[r][c] else [cr, cg, cb])

		# then blit the buffer
		# join not neccesarry when using a byte array 
		# buff =''.join(map(chr, buff))
		# update texture in UI thread.
		self._upd_texture(buff)

	def _upd_texture(self, buff):
		texture = self._qrtexture
		if not texture:
			# if texture hasn't yet been created delay the texture updation
			Clock.schedule_once(lambda dt: self._upd_texture(buff))
			return
		
		texture.blit_buffer(buff, colorfmt='rgb', bufferfmt='ubyte')
		texture.flip_vertical()
		img = self.qrimage
		img.anim_delay = -1
		img.texture = texture
		img.canvas.ask_update()

if __name__ == '__main__':
	from kivy.app import runTouchApp
	import sys
	data = str(sys.argv[1:])
	data = """{
	"host":"122.22.22.33",
	"port":12345
	}"""
	runTouchApp(Factory.QRCodeWidget(data=data))




from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.app import App
from kivy.clock import Clock
from kivy.factory import Factory
from kivy.utils import platform
from .baseWidget import VBox, HBox
from .toggleitems import PressableBox
from .utils import *
from .swipebehavior import SwipeBehavior

class PageContent(SwipeBehavior, VBox):
	def __init__(self, **kw):
		VBox.__init__(self, **kw)
		SwipeBehavior.__init__(self)

class MenuContainer(VBox):
	def __init__(self, **kw):
		VBox.__init__(self, **kw)
		self.bind(on_touch_down=self.check_press)
		self.register_event_type('on_dismiss')

	def on_dismiss(self, *args):
		print('on_dismiss fire')

	def open(self):
		if self in Window.children:
			Window.remove_widget(self)
		Window.add_widget(self)

	def dismiss(self, *args):
		self.clear_widgets()
		Window.remove_widget(self)
		self.dispatch('on_dismiss')

	def check_press(self, o, touch):
		if touch.is_mouse_scrolling:
			return False
		if not self.collide_point(touch.x, touch.y):
			self.dismiss()
			return True
		return False

class PagePanel(VBox):
	"""
# PagePanel widget
PagePanel widget provide a control bar and a subwidget container, control bar and content layout using vertical layout.

in control bar, there is a optional left menu icon, page title, right menu icon, and if there is more than one subwidgets add in the PagePanel, it will show the sub-widget last added. and a backward icon will show in the leftest control bar.

* backward icon uses to show the previous page in the PagePanel.
* left menu icon uses to show the system wide's menu
* right menu icon uses to show the current page's menu, the widget is identified by "menu_widget" in sub-widget description file
* title to show current page title, the widget is identified by "title_widget" in sub-widget description file.

## Description file format
PagePanel description file format
```
	{
		"bar_autohide": true when page is idle
		"bar_size": bar size in CSize unit
		"bar_at": "top" or "bottom"
		"bar_css":
		"panel_css":
		"left_menu": if defined, it must be a widget instance or a dict 
					recognized by Blocks

		other VBox initial options
	}
	usage examples:
	{
		"widgettype":"PagePanel",
		"options":{
			"bar_size":2,
			"enable_on_close":true,
			"left_menu":{
				"widgettype":"Text",
				"options":{
					"text":"Text"
				}
			},
			"radius":[10,10,10,10]
		},
		"subwidgets":[
			{
				"widgettype":"urlwidget",
				"options":{
					"url":"{{entire_url('page.ui')}}"
				}
			}
		]
	}
```

sub-widget's description file format
```
	and page.ui is:
	{
		"widgettype":"Button",
		"options":{
			"text":"button {{cnt or 0}}"
		},

		"title_widget":{
			"widgettype":"Text",
			"options":{
				"text":"test title"
			}
		},
		"menu_widget":{
			"widgettype":"Text",
			"options":{
				"text":"TTTT"
			}
		},
		"binds":[
			{
				"wid":"self",
				"target":"root",
				"actiontype":"urlwidget",
				"event":"on_press",
				"options":{
					"params":{
						"cnt":{{int(cnt or 0) + 1}}
					},
					"url":"{{entire_url('page.ui')}}"
				}
			}
		]
	}
```

## 

	"""
	def __init__(self, bar_size=2, 
					bar_css='default',
					csscls='default',
					singlepage=False,
					fixed_before=None,
					bar_autohide=False,
					fixed_after=None,
					bar_at='top', 
					enable_on_close=False, 
					left_menu=None, **kw):
		self.bar_size = bar_size
		self.bar_autohide = bar_autohide
		self.bar_at = bar_at
		self.singlepage = singlepage
		self.idle_status = False
		self.idle_threshold = 10
		self.bar_show = True
		self.idle_task = None
		self.swipe_buffer = []
		self.swipe_right = False
		self.fixed_before = None
		if fixed_before:
			self.fixed_before = Factory.Blocks().widgetBuild(fixed_before)
		self.fixed_after = None
		if fixed_after:
			self.fixed_after = Factory.Blocks().widgetBuild(fixed_after)
		
		self.enable_on_close = enable_on_close
		if self.enable_on_close:
			app = App.get_running_app()
			app.bind(on_close=self.on_close_handle)

		if not left_menu:
			self.left_menu = None
		elif isinstance(left_menu, Widget):
			self.left_menu = left_menu
		else:
			self.left_menu = Factory.Blocks().widgetBuild(left_menu)
		self.sub_widgets = []
		VBox.__init__(self, **kw)
		self.bar = HBox(size_hint_y=None,
						csscls=bar_css,
						spacing=CSize(bar_size/6),
						height=CSize(bar_size))
		bcsize = bar_size * 0.85
		self.content = PageContent()
		self.content.bind(on_swipe_left=self.on_swipe_next_page)
		self.content.bind(on_swipe_right=self.pop)
		self.bar_back = VBox(size_hint=(None,None),size=CSize(bcsize,bcsize))
		self.bar_back_w = Factory.Blocks().widgetBuild({
			"widgettype":"PressableBox",
			"options":{
				"size":CSize(bcsize,bcsize),
				"size_hint":[None, None]
			},
			"subwidgets":[
				{
					"widgettype":"AsyncImage",
					"options":{
						"source":blockImage('backword.png')
					}
				}
			]
		})
		self.bar_back_w.bind(on_press=self.pop)
		self.bar.add_widget(self.bar_back)
		if self.left_menu:
			self.bar_left_menu = Factory.Blocks().widgetBuild({
				"widgettype":"PressableBox",
				"options":{
					"size_hint":(None,None), 
					"size":CSize(bcsize,bcsize)
				},
				"subwidgets":[
					{
						"widgettype":"AsyncImage",
						"options":{
							"source":blockImage('menu.png')
						}
					}
				]
			})
			self.bar.add_widget(self.bar_left_menu)
			self.bar_left_menu.bind(on_press=self.show_left_menu)
		self.bar_title = HBox(csscls=bar_css)
		self.bar.add_widget(self.bar_title)
		self.bar_right_menu = VBox(size_hint=(None,None),size=CSize(bcsize,bcsize))
		self.bar_right_menu_w = Factory.Blocks().widgetBuild({
			"widgettype":"PressableBox",
			"options":{
				"size_hint":(None,None), 
				"size":CSize(bcsize,bcsize)
			},
			"subwidgets":[
				{
					"widgettype":"AsyncImage",
					"options":{
						"source":blockImage('right_menu.png')
					}
				}
			]
		})
		self.bar.add_widget(self.bar_right_menu)
		self.bar_right_menu_w.bind(on_press=self.show_right_menu)
		self.construct()
		if self.bar_autohide:
			Window.bind(on_touch_down=self.set_normal_bar)
			self.idle_task = Clock.schedule_once(self.set_idle_bar, \
													self.idle_threshold)

	def set_idle_bar(self, *args):
		if not self.bar_show:
			return
		try:
			self.bar_pos = self.children.index(self.bar)
			print('self.bar_pos=', self.bar_pos, '......................')
			super().remove_widget(self.bar)
			if platform in ['win', 'macosx','linux']:
				Window.borderless = True
		except:
			pass
		self.bar_show = False

	def set_normal_bar(self, *args):
		if self.idle_task:
			self.idle_task.cancel()
		self.idle_task = Clock.schedule_once(self.set_idle_bar, \
													self.idle_threshold)
		if self.bar_show:
			return
		super().add_widget(self.bar, index=self.bar_pos)
		if platform in ['win', 'macosx','linux']:
			Window.borderless = False
		self.bar_show = True

	def construct(self):
		self.clear_widgets()
		if self.bar_show and self.bar_at == 'top':
			super().add_widget(self.bar)
		if self.fixed_before:
			super().add_widget(self.fixed_before)
		super().add_widget(self.content)
		if self.fixed_after:
			super().add_widget(self.fixed_after)
		if self.bar_show and self.bar_at != 'top':
			super().add_widget(self.bar)
		self.left_menu_showed = False
		self.right_menu_showed = False

	def get_subwidgets(self):
		children = [self.bar]
		if self.fixed_before:
			children.append(self.fixed_before)
		if self.fixed_after:
			children.append(self.fixed_after)
		return children + self.sub_widgets

	def on_close_handle(self, o, *args):
		print('app.on_close fired, ...')
		if not self.enable_on_close:
			return True
		self.pop(None)
		if len(self.sub_widgets) > 1:
			return False
		return True

	def pop(self, o, *args):
		if len(self.sub_widgets) < 2:
			return
		self.clear_widgets()
		diss_w = self.sub_widgets[-1]
		self.swipe_buffer.insert(0,diss_w)
		self.sub_widgets = self.sub_widgets[:-1]
		self.show_currentpage()

	def set_title(self, w):
		self.bar_title.clear_widgets()
		self.bar_title.add_widget(w)

	def set_right_menu(self, w):
		self.bar_right_menu.clear_widgets()
		self.bar_right_menu.add_widget(w)

	def show_currentpage(self):
		w = self.sub_widgets[-1]
		if len(self.sub_widgets) > 1:
			self.bar_back.add_widget(self.bar_back_w)
		self.content.add_widget(w)
		if hasattr(w, 'title_widget'):
			self.set_title(w.title_widget)
		if hasattr(w, 'menu_widget'):
			self.set_right_menu(self.bar_right_menu_w)

	def on_swipe_next_page(self, o, *args):
		if len(self.swipe_buffer) < 1:
			return True
		self.swipe_right = True
		w = self.swipe_buffer[0]
		del self.swipe_buffer[0]
		self.add_widget(w)
		return True

	def clear_widgets(self):
		self.bar_back.clear_widgets()
		self.content.clear_widgets()
		self.bar_right_menu.clear_widgets()

	def add_widget(self, w, *args):
		print('here ....')
		if not self.swipe_right:
			self.swipe_buffer = []
		self.swipe_right = False
		self.clear_widgets()
		if self.singlepage:
			self.sub_widgets = []
		self.sub_widgets.append(w)
		self.show_currentpage()

	def show_left_menu(self, o):
		def x(*args):
			self.left_menu_showed = False
			print('dismiss fired, left_menu_showed=',self.left_menu_showed)

		print('left_menu fired')
		if len(self.sub_widgets) < 1:
			return
		if self.left_menu_showed:
			return
		mc = MenuContainer()
		mc.add_widget(self.left_menu)
		self.left_menu.bind(on_press=mc.dismiss)
		mc.size_hint = (None, None)
		mc.width = self.width * 0.4
		mc.height = self.content.height
		mc.x = self.x
		mc.y = self.y if self.bar_at=='top' else self.content.y
		self.left_menu_showed = True
		mc.bind(on_dismiss=x)
		mc.open()

	def show_right_menu(self, o):
		def x(*args):
			self.right_menu_showed = False
			print('dismiss fired, right_menu_showed=',self.right_menu_showed)

		print('right menu fired')
		if len(self.sub_widgets) < 1:
			print('no sub_widgets')
			return
		if self.right_menu_showed:
			print('right menu showed, not thing done')
			return

		w = self.sub_widgets[-1]
		if not hasattr(w, 'menu_widget'):
			print('this sub widget has not menu_widget')
			return True
		mc = MenuContainer()
		mc.add_widget(w.menu_widget)
		w.menu_widget.bind(on_press=mc.dismiss)
		mc.size_hint = (None,None)
		mc.height = self.content.height
		mc.width = self.width * 0.4
		mc.x = self.x + self.width * 0.6
		mc.y = self.y if self.bar_at == 'top' else self.content.y
		self.right_menu_showed = True
		mc.bind(on_dismiss=x)
		mc.open()
		



from kivy.uix.image import Image,AsyncImage
from kivy.uix.button import Button
from kivy.graphics import Color
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import NumericProperty

class BinStateImage(Button):
	binstate = NumericProperty(1)
	def __init__(self,source1='../images/bullet_arrow_right.png',source2='../images/bullet_arrow_down.png',**kv):
		self.bin_state_images = [
			source1,
			source2
		]
		kv.update({'source':source1})
		super(BinStateImage,self).__init__()
		self.myImgWidget = Image(**kv)
		self.size = self.myImgWidget.size
		#self.background_color = Color(1,0,0,1)
		self.bind(on_release=self.changeState)
	
	def changeState(self,instance):
		#print('on_release fired')
		if instance.binstate == 1:
			instance.binstate = 2
		else: instance.binstate = 1
		self.myImgWidget.source = self.bin_state_images[instance.binstate-1]
		self.myImgWidget.reload()

if __name__ == '__main__':
	from kivy.app import App
	class MyApp(App):
		def build(self):
			m = BoxLayout(orientation='vertical')
			img = BinStateImage()
			img.bind(binstate=self.showState)
			m.add_widget(img)
			b = Button(text='GGGG')
			m.add_widget(b)
			return m
		def showState(self,img,v):
			pass #print('cur state=',img.binstate,v)
	
	MyApp().run()




from kivy.event import EventDispatcher
from kivy.app import App
from kivy.factory import Factory
from .threadcall import HttpClient
from .utils import absurl
from appPublic.registerfunction import RegisterFunction

class DataGraber(EventDispatcher):
	"""
	Graber format
	{
		"widgettype":"DataGraber",
		"options":{
			"dataurl":"first",
			"datarfname":"second",
			"target":"third",
			"params":
			"method":
			"pagging":"default false"
		}
	}
	if dataurl present, the DataGraber using this dataurl to get and 
		return data
	else if datarfname present, it find a registered function named 
		by 'datarfname' to return data
	else if datatarget present, it find the widget and uses target
		method(default is 'getValue') to return data
	else it return None
	"""
	def __init__(self, **kw):
		super().__init__()
		self.options = kw
		self.register_event_type('on_success')
		self.register_event_type('on_error')

	def load(self, *args, **kw):
		ret = None
		while True:
			try:
				dataurl = self.options.get('dataurl')
				if dataurl:
					ret = self.loadUrlData(*args, **kw)
					break

				rfname = self.options.get('datarfname')
				if rfname:
					ret = self.loadRFData(*args, **kw)
					break
				target = self.options.get('datatarget')
				if target:
					ret = self.loadTargetData(*args, **kw)
					break
			except Exception as e:
				self.dispatch('on_error', e)
				return
		if ret:
			self.dispatch('on_success',ret)
		else:
			e = Exception('Not method to do load')
			self.dispatch('on_error', e)

	def loadUrlData(self, *args, **kw):
		dataurl = self.options.get('dataurl')
		hc = HttpClient()
		params = self.options.get('params',{}).copy()
		params.update(kw)
		method = self.options.get('method','GET')
		d = hc(dataurl, params=params,method=method)
		return d

	def loadRFData(self, *args, **kw):
		rfname = self.options.get('datarfname')
		rf = RegisterFunction()
		f = rf.get(rfname)
		if not f:
			return None
		params = self.options.get('params',{}).copy()
		params.update(kw)
		try:
			d = f(**params)
			return d
		except Exception as e:
			Logger.info('blocks : Exception:%s',e)
			print_exc()
		return None

	def loadTargetData(self, *args, **kw):
		target = self.options.get('datatarget')
		w = Factory.Blocks.getWidgetById(target)
		if not w:
			return None
		params = self.options.get('params',{}).copy()
		params.update(kw)
		method = params.get('method', 'getValue')
		if not has(w, method):
			return None
		try:
			f = getattr(w, method)
			d = f()
			return d
		except Exception as e:
			Logger.info('blocks : Exception %s', e)
			print_exc()
		return None

	def on_success(self,d):
		pass

	def on_error(self,e):
		pass
	
class DataLoader(EventDispatcher):
	def __init__(self,data_user):
		self.data_user = data_user
		EventDispatcher.__init__(self)
		self.register_event_type('on_success')
		self.register_event_type('on_error')

	def on_success(self,d):
		pass

	def on_error(self,e):
		pass
	
	def load(self):
		pass

class HttpDataLoader(DataLoader):
	def load(self, *args, **kw):
		app = App.get_running_app()
		url = app.realurl(self.data_user.url)
		method = self.data_user.method
		params = self.data_user.params.copy()
		params.update({
			"page":self.data_user.curpage,
			"rows":self.data_user.page_rows
		})
		hc = HttpClient()
		try:
			r = hc(url, method=method, params=params)
			self.dispatch('on_success', r)
			return r
		except Exception as e:
			self.dispatch('on_error', e)


class ListDataLoader(DataLoader):
	def load(self, *args, **kw):
		p = self.data_user.curpage
		r = self.data_user.page_rows
		try:
			s = self.data_user.data[(p-1)*r:p*r]
			d = {
				"total":len(self.data_user.data),
				"rows":s
			}
			self.dispatch('on_success', d)
			return d
		except Exception as e:
			self.dispatch('on_error', e)

class RegisterFunctionDataLoader(DataLoader):
	def load(self, *args, **kw):
		rf = RegisterFunction()
		try:
			rfname = self.data_user.rfname
			func = rf.get(rfname)
			if func is None:
				raise Exception('%s is not a registerfunction' % rfname)
			params = {k:v for k,v in self.user_data.params.items()}
			params.update({
				"page":self.data_user.curpage,
				"rows":self.data_user.page_rows
			})
			s = func(**params)
			self.dispatch('on_success', s)
			return s
		except Exception as e:
			self.dispatch('on_error', e)





# coding=utf-8
"""
MapView
=======

MapView is a Kivy widget that display maps.
"""
from kivyblocks.mapview.source import MapSource
from kivyblocks.mapview.types import Bbox, Coordinate
from kivyblocks.mapview.view import (
    MapLayer,
    MapMarker,
    MapMarkerPopup,
    MapView,
    MarkerMapLayer,
)

__all__ = [
    "Coordinate",
    "Bbox",
    "MapView",
    "MapSource",
    "MapMarker",
    "MapLayer",
    "MarkerMapLayer",
    "MapMarkerPopup",
]



# from appPublic.background import Background
from threading import Thread
from kivy.event import EventDispatcher

class BackendFunction(EventDispatcher, Thread):
	def __init__(self, callee, *args, **kw):
		EventDispatcher.__init__(self)
		Thread.__init__(self)
		self._callee = callee
		self._args = args
		self._kw = kw
		self.register_event_type('on_success')
		self.register_event_type('on_failed')

	def on_success(self, ret):
		print('BackendFunction(), on_success fired')

	def on_failed(self, e):
		print('BackendFunciton(), on_failed fired', e)

	def run(self):
		try:
			x = self._callee(*self._args, **self._kw)
			self.dispatch('on_success', x)
		except Exception as e:
			self.dispatch('on_failed',e)



import os
import traceback
try:
	import ujson as json
except:
	import json
from os.path import join, sep, dirname, basename
from appPublic.Singleton import GlobalEnv
from appPublic.jsonConfig import getConfig
from appPublic.myTE import MyTemplateEngine
from appPublic.Singleton import SingletonDecorator
from appPublic.dictObject import DictObject

from kivy.logger import Logger
from kivy.utils import platform

@SingletonDecorator
class ScriptEnv(DictObject):
	pass

def set_script_env(n,v):
	env = ScriptEnv()
	env.update({n:v})

class Script:
	def __init__(self):
		config = getConfig()
		self.root = config.script_root
		print('Script.root=', self.root)
		self.env = {}
		self.handlers = {}
		self.register('.tmpl', TemplateHandler)
		self.register('.dspy', DspyHandler)
		self.register('.ui', TemplateHandler)

	def url2filepath(self, url):
		if url.startswith('file://'):
			url = url[7:]
		return join(self.root, *url.split('/'))

	def show_info(self, env):
		workdir = env['workdir']
		sfile = env['filepath']
		url = env['url']
		print(f'script:workdir={workdir}')
		print(f'script:script_file={sfile}')
		print(f'script:url={url}')
		sdir = os.path.join(workdir, 'scripts')
		sf_exists = os.path.isdir(sdir)
		conf_f = os.path.join(workdir, 'conf', 'config.json')
		conf_exists = os.path.isfile(conf_f)
		print(f'script:script exists {sf_exists}')
		print(f'script:config.json exists {conf_exists}')

	def dispatch(self, url, **kw):
		filepath = self.url2filepath(url)
		for suffix, handler in self.handlers.items():
			if filepath.endswith(suffix):
				env = self.env.copy()
				env.update(ScriptEnv())
				env.update(kw)
				env['root_path'] = self.root
				env['url'] = url
				env['filepath'] = filepath
				print(f"workdir={env['workdir']}--------")
				if platform == 'android':
					self.show_info(env)

				h = handler(env)
				d = h.render()
				try:
					return json.loads(d)
				except:
					return d

	def register(self, suffix, handler):
		self.handlers[suffix] = handler

	def set_env(self, n, v):
		self.env.update({n:v})

class BaseHandler:
	def entire_url(self, url):
		if url.startswith('file://') or \
				url.startswith('http://') or \
				url.startswith('hppts://'):
			return url
		tokens = url.split('/')
		if tokens[0] == '':
			root = self.env['root_path']
			tokens[0] = root
			return '/'.join(tokens)

		p1 = self.env['url'].split('/')[:-1]
		return '/'.join(p1+tokens)

	def __init__(self, env):
		self.env = env
		self.env['entire_url'] = self.entire_url

class TemplateHandler(BaseHandler):
	def __init__(self, env):
		super().__init__(env)
		root = env['root_path']
		paths = [root]
		fp = env['filepath'][len(root):]
		plst = fp.split(sep)[:-1]
		self.templ_file = basename(env['filepath'])
		cpath = root
		for p in plst:
			cpath = join(cpath, p)
			paths.append(cpath)

		paths.reverse()
		self.engine = MyTemplateEngine(paths)

	def render(self):
		try:
			return self.engine.render(self.templ_file, self.env)
		except Exception as e:
			print('Exception:', str(e))
			print('filename=', self.env['filepath'])
			traceback.print_exc()

class DspyHandler(BaseHandler):
	def __init__(self, env):
		super().__init__(env)

	def loadScript(self, path):
		data = ''
		with codecs.open(path,'rb','utf-8') as f:
			data = f.read()
		b= ''.join(data.split('\r'))
		lines = b.split('\n')
		lines = ['\t' + l for l in lines ]
		txt = "def myfunc(request,**ns):\n" + '\n'.join(lines)
		return txt

	def render(self, params={}):
		try:
			lenv = self.env.copy()
			lenv.update(params)
			txt = self.loadScript(self.env['filepath'])
			exec(txt,lenv,lenv)
			func = lenv['myfunc']
			return func(self.env, **lenv)
		except Exception as e:
			print('Exception:', str(e))
			print('filename=', self.env['filepath'])
			traceback.print_exc()




from functools import partial
from kivy.logger import Logger
from kivy.uix.behaviors import TouchRippleButtonBehavior
from kivy.graphics import Color, Rectangle
from kivy.uix.boxlayout import BoxLayout
from kivy.factory import Factory
from kivy.properties import ObjectProperty, StringProperty, BooleanProperty

from kivyblocks.ready import WidgetReady
from kivyblocks.bgcolorbehavior import BGColorBehavior
from kivyblocks.utils import CSize
from kivyblocks.baseWidget import Box
from kivyblocks.widget_css import WidgetCSS

class PressableBox(TouchRippleButtonBehavior, Box):
	normal_css = StringProperty("default")
	actived_css = StringProperty("default")
	box_actived = BooleanProperty(False)
	def __init__(self,
				border_width=1,
				user_data=None,
				radius=[],
				**kw):
		super(PressableBox, self).__init__(
			padding=[border_width,
			border_width,
			border_width,
			border_width],
			radius=radius,
			**kw)
		self.border_width = border_width
		self.user_data = user_data
		self.actived = False
		self.csscls = self.normal_css

	def active(self, flag):
		self.box_actived = flag

	def on_box_actived(self, o, v):
		if self.box_actived:
			self.csscls = self.actived_css
		else:
			self.csscls = self.normal_css

	def on_press(self,o=None):
		self.box_actived = True


	def setValue(self,d):
		self.user_data = d

	def getValue(self):
		return self.user_data

"""
ToggleItems format:
{
	radius:
	unit_size:
	items_desc:
	border_width:
	orientation:
}
"""
class ToggleItems(BoxLayout, WidgetCSS):
	def __init__(self,
				radius=[],
				orientation='horizontal',
				unit_size=3,
				items_desc=[],
				border_width=1,
				normal_css="default",
				actived_css="default",
				**kw):
		self.unit_size_pix = CSize(unit_size) 
		kw1 = {
			"orientation":orientation
		}
		kw1.update(kw)
		if orientation=='horizontal':
			kw1['size_hint_y'] = None
			kw1['height'] = self.unit_size_pix
			kw1['size_hint_min_x'] = self.unit_size_pix
		else:
			kw1['size_hint_x'] = None
			kw1['width'] = self.unit_size_pix
			kw1['size_hint_min_y'] = self.unit_size_pix
		
		BoxLayout.__init__(self, **kw1)
		self.item_widgets = []
		kw = {
			"border_width":border_width,
			"normal_css":normal_css,
			"actived_css":actived_css,
			"radius":radius
		}
		kw.update(kw1)

		for desc in items_desc:
			c = PressableBox(**kw)
			self.item_widgets.append(c)
			self.add_widget(c)
			c.bind(on_press=self.select_item)
			if desc.get('data'):
				c.setValue(desc['data'])

			b = Factory.Blocks()
			def cb(c,o,w):
				c.add_widget(w)
			def eb(desc,o,e):
				print(desc,'error',e)
			b.bind(on_built=partial(cb,c))
			b.bind(on_failed=partial(eb,desc))
			b.widgetBuild(desc)

		if len(self.item_widgets) > 0:
			self.item_widgets[0].actived = True
		else:
			print('items_desc=',items_desc,'error')
		self.register_event_type('on_press')
	
	def on_press(self,v=None):
		print('Toggle on_press fired')
		pass

	def select_item(self,o=None):
		for i in self.item_widgets:
			if i!=o:
				i.box_actived = False

		self.dispatch('on_press',o.getValue())

	def getValue(self):
		for i in self.item_widgets:
			if i.actived:
				return i.getValue()
		return None

	def setValue(self,v):
		for i in self.item_widgets:
			if i.getValue() == v:
				i.actived = True
				self.select_item(i)
			else:
				i.actived = False




# -*- coding: utf-8 -*-
from kivy import platform
from kivy.core.window import Window
from kivy.metrics import dp
from kivyblocks import fonts_path

# Feel free to override this const if you're designing for a device such as
# a GNU/Linux tablet.
if platform != "android" and platform != "ios":
    DEVICE_TYPE = "desktop"
elif Window.width >= dp(600) and Window.height >= dp(600):
    DEVICE_TYPE = "tablet"
else:
    DEVICE_TYPE = "mobile"

if DEVICE_TYPE == "mobile":
    MAX_NAV_DRAWER_WIDTH = dp(300)
    HORIZ_MARGINS = dp(16)
    STANDARD_INCREMENT = dp(56)
    PORTRAIT_TOOLBAR_HEIGHT = STANDARD_INCREMENT
    LANDSCAPE_TOOLBAR_HEIGHT = STANDARD_INCREMENT - dp(8)
else:
    MAX_NAV_DRAWER_WIDTH = dp(400)
    HORIZ_MARGINS = dp(24)
    STANDARD_INCREMENT = dp(64)
    PORTRAIT_TOOLBAR_HEIGHT = STANDARD_INCREMENT
    LANDSCAPE_TOOLBAR_HEIGHT = STANDARD_INCREMENT

TOUCH_TARGET_HEIGHT = dp(48)

FONTS = [
    {
        "name": "Roboto",
        "fn_regular": fonts_path + 'Roboto-Regular.ttf',
        "fn_bold": fonts_path + 'Roboto-Medium.ttf',
        "fn_italic": fonts_path + 'Roboto-Italic.ttf',
        "fn_bolditalic": fonts_path + 'Roboto-MediumItalic.ttf'
    },
    {
        "name": "RobotoLight",
        "fn_regular": fonts_path + 'Roboto-Thin.ttf',
        "fn_bold": fonts_path + 'Roboto-Light.ttf',
        "fn_italic": fonts_path + 'Roboto-ThinItalic.ttf',
        "fn_bolditalic": fonts_path + 'Roboto-LightItalic.ttf'
    },
    {
        "name": "Icons",
        "fn_regular": fonts_path + 'materialdesignicons-webfont.ttf'
    }
]



from kivy.clock import Clock
from kivy.factory import Factory
from kivy.uix.stacklayout import StackLayout
from kivy.properties import BooleanProperty, StringProperty,\
		ListProperty, DictProperty, OptionProperty, \
		NumericProperty

from kivyblocks.baseWidget import VBox 
from kivyblocks.clickable import ToggleText, ToggleImage
from kivyblocks.utils import SUPER, CSize
from kivyblocks.widget_css import WidgetCSS
from kivyblocks.ready import WidgetReady

class MultiSelect(WidgetCSS, WidgetReady, StackLayout):
	"""
	{
		"widgettype":"MultiSelect",
		"options":{
			"valueField"
			"textField"
			"items"
			"item_css"
			"item_selected_css"
			"all_position"
			"need_others"
			"default_selected"
		}
	}
	"""
	valueField = StringProperty('value')
	textField = StringProperty('text')
	x_spacing_c = NumericProperty(1)
	y_spacing_c = NumericProperty(0.5)
	items = ListProperty([])
	item_css = StringProperty({})
	item_selected_css = StringProperty({})
	all_position = OptionProperty(None, \
					options=[None, 'begin', 'end'])
	need_others = BooleanProperty(False)
	default_selected=ListProperty([])
	result = ListProperty([])
	def __init__(self, **kw):
		SUPER(MultiSelect, self, kw)
		self.spacing = CSize(self.x_spacing_c, self.y_spacing_c)
		self.orientation = 'lr-tb'
		self.build_all_items()
		self.old_result = []
		self.register_event_type('on_changed')
		self.button_dic = {}
		self.value_b = {}
		self.buildWidgets()
	
	def build_all_items(self):
		items = [i for i in self.items ]
		if self.need_others:
			dic = {
				self.valueField:'_other_',
				self.textField:'others'
			}
			items.append(dic)

		if self.all_position is not None:
			dic = {
				self.textField:'all',
				self.valueField:'_all_',
			}
			if self.all_position == 'begin':
				items.insert(0, dic)
			else:
				items.append(dic)
		self.all_items = items

	def on_changed(self, d):
		pass

	def build_item_widget(self, dic):
		keys = dic.keys()
		ToggleText = Factory.ToggleText
		ToggleImage = Factory.ToggleImage
		ToggleIconText = Factory.ToggleIconText
		if self.textField in keys and 'source_on' in keys and \
					'source_off' in keys:
			desc['text'] = dic[self.textField]
			desc['css_off'] = self.item_css
			desc['css_on'] = self.item_selected_css
			desc["source_on"] = dic['source_on'],
			desc["source_off"] = dic['source_off']
			print('ToggleIconText', desc)
			return ToggleIconText(**desc)
			
		if self.textField in keys:
			desc = {}
			desc['text'] = dic[self.textField]
			desc['css_off'] = self.item_css
			desc['css_on'] = self.item_selected_css
			print('ToggleText', desc)
			return ToggleText(**desc)
		if 'source_on' in keys and 'source_off' in keys:
			desc = {
				"source_on":dic['source_on'],
				"source_off":dic['source_off']
			}
			print('ToggleImage', desc)
			return ToggleImage(**desc)

	def buildWidgets(self):
		self.clear_widgets()
		for item in self.all_items:
			b = self.build_item_widget(item)
			b.bind(on_press=self.button_pressed)
			self.button_dic.update({b:item})
			self.add_widget(b)

	def button_pressed(self, o):
		o.toggle()
		if self.button_dic[o][self.valueField] == '_all_' and o.state():
			for b, dic in self.button_dic.items():
				if b == o:
					continue
				b.select(False)
		if self.button_dic[o][self.valueField] == '_other_' and o.state():
			for b, dic in self.button_dic.items():
				if b == o:
					continue
				b.select(False)
		if self.button_dic[o][self.valueField] not in ['_all_', '_other_'] \
						and o.state():
			for b, dic in self.button_dic.items():
				if dic[self.valueField] in ['_all_', '_other_']:
					b.select(False)
		Clock.schedule_once(self.set_result, 0.1)

	def set_result(self, t=None):
		self.result = [ d[self.valueField] for b, d in \
							self.button_dic.items() if b.state() ]
		if self.result != self.old_result:
			self.dispatch('on_changed', self.result)
		self.old_result = self.result
		
	def getValue(self):
		self.set_result()
		return self.result

	def setValue(self, value):
		self.result = value
		self.old_result = value
		for v in value:
			b = self.value_b[v]
			b.select(True)






from kivy.factory import Factory
from kivy.core.window import Window
from kivy.properties import StringProperty, NumericProperty, \
		OptionProperty
from kivy.uix.image import AsyncImage
from .utils import *
from .baseWidget import Modal, Text, HBox,VBox, TimedModal
from .widget_css import register_css
from .clickable import ClickableIconText
from .toggleitems import PressableBox

error_default_title_css = {
	"bgcolor":[0.88,0,0,1],
	"fgcolor":[0.12,0.12,0.12,1]
}

conform_default_title_css = {
	"bgcolor":[0.88,0.88,0.88,1],
	"fgcolor":[0.12,0.12,0.12,1]
}
conform_default_body_css = {
	"bgcolor":[0.92,0.92,0.92,1],
	"fgcolor":[0.16,0.16,0.16,1]
}

conform_button_default_css = {
	"bgcolor":[0.05,0.85,0.05,1],
	"fgcolor":[0.95,0.95,0.95,1]
}

cancel_button_default_css = {
	"bgcolor":[0.85, 0.5, 0.5, 1],
	"fgcolor":[0.9, 0.90, 0.9, 1]
}

register_css('conform_default_title_css', conform_default_title_css)
register_css('message_default_title_css', conform_default_title_css)
register_css('error_default_title_css', error_default_title_css)
register_css('conform_default_body_css', conform_default_body_css)
register_css('message_default_body_css', conform_default_body_css)
register_css('conform_button_default_css', conform_button_default_css)
register_css('cancel_button_default_css', cancel_button_default_css)

class Message(TimedModal):
	title = StringProperty(None)
	default_title = 'Message'
	default_message = 'Message'
	title_icon = StringProperty(blockImage('info.png'))
	message = StringProperty(None)
	body_css = StringProperty('message_default_body_css')
	title_css = StringProperty('message_default_title_css')
	show_position = OptionProperty('br', options=['tl','tc','tr',
													'cl','cc','cr',
													'bl','bc','br'])
	width_c = NumericProperty(15)
	height_c = NumericProperty(6)
	def __init__(self, **kw):
		SUPER(Message, self, kw)
		self.size_hint = None, None
		self.set_position()
		b = VBox(csscls=self.title_css)
		b1 = HBox(size_hint_y=None, height=CSize(2))
		b1.add_widget(AsyncImage(source=self.title_icon,
									size_hint=[None,None],
									height=CSize(1.2),
									width=CSize(1.2)))
		b1.add_widget(Text(otext=self.title or self.default_title,
									font_size=CSize(1.2),
									i18n=True,
									wrap=True,
									halign='left',
									valign='center',
									))
		b.add_widget(b1)
		b2 = HBox(csscls=self.body_css)
		b2.add_widget(Text(text=self.message or self.default_message,
									i18n=True,
									wrap=True,
									halign='left',
									size_hint=(1,1),
									font_size=CSize(1)))
		b.add_widget(b2)
		self.add_widget(b)
	
	def on_size(self, *args):
		self.set_position()
		try:
			super().on_size(*args)
		except:
			pass

	def set_position(self):
		# self.pos_hint = None, None
		self.size_hint = None, None
		self.width = CSize(self.width_c)
		self.height = CSize(self.height_c)
		xn = self.show_position[1]
		yn = self.show_position[0]

		if xn == 'l':	
			self.anchor_x = 'left'
		elif xn == 'c':
			self.anchor_x = 'center'
		else:
			self.anchor_x = 'right'
		if yn == 't':
			self.anchor_y = 'top'
		elif yn == 'c':
			self.anchor_y = 'center'
		else:
			self.anchor_y = 'bottom'

class Error(Message):
	default_title = 'Error'
	default_message = 'Error message'
	title_icon = StringProperty(blockImage('error.png'))
	title_css = StringProperty('error_default_title_css')

class Conform(Modal):
	title = StringProperty(None)
	title_icon = StringProperty(None)
	message = StringProperty(None)
	conform_icon = StringProperty(None)
	cancel_icon = StringProperty(None)
	title_css = StringProperty('conform_default_title_css')
	body_css = StringProperty('conform_default_body_css')
	conform_css = StringProperty('conform_button_default_css')
	cancel_css = StringProperty('cancel_button_default_css')
	def __init__(self, **kw):
		super(Conform, self).__init__(**kw)
		self.register_event_type('on_cancel')
		self.register_event_type('on_conform')
		self.auto_dismiss = False
		b = VBox(csscls=self.title_css)
		b1 = HBox(size_hint_y=None, height=CSize(2))
		b1.add_widget(AsyncImage(source=self.title_icon or \
									blockImage('question.png'),
									size_hint=[None,None],
									height=CSize(1.2),
									width=CSize(1.2)))
		b1.add_widget(Text(otext=self.title or 'Conform',
									font_size=CSize(1.2),
									i18n=True,
									wrap=True,
									halign='left',
									valign='center',
									))
		b.add_widget(b1)
		b2 = HBox(csscls=self.body_css)
		
		b2.add_widget(Text(text=self.message or 'Please conform',
									i18n=True,
									wrap=True,
									halign='left',
									size_hint=(1,1),
									font_size=CSize(1)))
		b.add_widget(b2)
		b3 = HBox(size_hint_y=None, 
					height=CSize(2), 
					csscls=self.body_css)
		w_cancel = ClickableIconText(text='Cancel',
							source=self.cancel_icon or
									blockImage('cancel.png'),
							img_kw={
								"size_hint":[None, None],
								"height":CSize(1),
								"width":CSize(1)
							},
							csscls=self.cancel_css
							)
		w_conform = ClickableIconText(text='Conform',
							source=self.conform_icon or
									blockImage('question.png'),
							img_kw={
								"size_hint":[None, None],
								"height":CSize(1),
								"width":CSize(1)
							},
							csscls=self.conform_css
							)
		w_cancel.bind(on_press=self.cancel_press)
		w_conform.bind(on_press=self.conform_press)

		b3.add_widget(w_conform)
		b3.add_widget(w_cancel)
		b.add_widget(b3)
		self.add_widget(b)

	def cancel_press(self,o, v=None):
		self.dispatch('on_cancel')

	def conform_press(self, o, v=None):
		self.dispatch('on_conform')

	def on_cancel(self):
		self.dismiss()

	def on_conform(self):
		self.dismiss()





from traceback import print_exc
import json
from kivy.properties import StringProperty
from kivy.factory import Factory
from .toolbar import Toolbar
from .baseWidget import VBox, StrInput
from .utils import SUPER, blockImage, CSize
from .clickable import ClickableText

class BlockTest(VBox):
	source_code = StringProperty(None)
	def __init__(self, **kw):
		SUPER(BlockTest, self, kw)
		tb_desc = {
			"img_size_c":2,
			"toolbar_orient":"H",
			"tool_orient":"horizontal",
			"css_on":"default",
			"css_off":"default",
			"tools":[
				{
					"name":"source",
					"source_on":blockImage('source_on.png'),
					"source_off":blockImage('source_off.png'),
					"label":"source code"
				},
				{
					"name":"widget",
					"source_on":blockImage('widget_on.png'),
					"source_off":blockImage('widget_off.png'),
					"label":"Result widget"
				}
			]
		}
		self.toolbar = Toolbar(**tb_desc)
		self.toolbar.bind(on_press=self.tool_pressed)
		self.content = VBox()
		self.source_w = StrInput(multiline=True)
		self.content.add_widget(self.source_w)
		if self.source_code:
			self.source_w.text = self.source_code
		self.add_widget(self.toolbar)
		self.add_widget(self.content)

	def tool_pressed(self, o, v):
		self.content.clear_widgets()
		if v['name'] == 'source':
			print('switch to source code')
			self.content.add_widget(self.source_w)
		else:
			try:
				dic = json.loads(self.source_w.text)
				print(dic, type(dic))
				w = Factory.Blocks().widgetBuild(dic)
				self.content.add_widget(w)
			except Exception as e:
				print_exc()
				print('BlockTest:Exception:',e)

Factory.register('BlockTest', BlockTest)



import time
from kivy.uix.video import Video
from kivy.logger import Logger
from kivy.core.window import Window
from kivy.factory import Factory
from kivy.properties import BooleanProperty, DictProperty
from kivy.clock import Clock

from ffpyplayer.tools import set_log_callback

from .hack_ffpyplayer import set_headers_pattern, hack_mediaplayer
logger_func = {'quiet': Logger.critical, 'panic': Logger.critical,
               'fatal': Logger.critical, 'error': Logger.error,
               'warning': Logger.warning, 'info': Logger.info,
               'verbose': Logger.debug, 'debug': Logger.debug}


class NewVideo(Video):
	center_screen = BooleanProperty(False)
	h_pattern = DictProperty(None)
	def __init__(self, **kw):
		hack_mediaplayer()
		Video.__init__(self, **kw)
		Window.allow_screensaver = False
		set_log_callback(self.ffplayerLog)
		self.register_event_type('on_open_failed')
		self.register_event_type('on_leave_focus')
		self.register_event_type('on_enter_focus')
		self.register_event_type('on_load_success')
		self.bind(source=self.record_start_time)
		self.bind(loaded=self.on_video_loaded)
		self.load_status = None
		Clock.schedule_interval(self.check_focus,0.2)
		if self.h_pattern:
			[ set_headers_pattern(k,v) for k,v in self.h_pattern.items() ]

	def set_patternheaders(self, pattern, headers_str):
		set_headers_pattern(pattern, headers_str)

	def on_enter_focus(self, v=None):
		pass

	def on_leave_focus(self, v=None):
		pass

	def check_focus(self,*args):
		if not self.parent:
			self.center_screen = False
			return
		if not self.get_root_window():
			self.center_screen = False

		w = self.parent
		pos = w.to_widget(*Window.center)
		if self.collide_point(*pos):
			self.center_screen = True
		else:
			self.center_screen = False

	def on_center_screen(self,o, v):
		if v:
			self.dispatch('on_enter_focus')
		else:
			self.dispatch('on_leave_focus')

	def resize(self, *args):
		v_size = self.get_video_size()
		if v_size:
			self.height = self.width * v_size[1] / v_size[0]

	def on_load_success(self, *args):
		pass

	def record_start_time(self, *args):
		self.start_time = time.time()
		self.load_status = 'start'

	def on_open_failed(self, source, x=None):
		Logger.error(f'NewVideo: source={self.source} open failed')
		self.load_status = 'failed'

	def ffplayerLog(self, msg, level):
		if 'Connection to tcp' in msg and 'failed' in msg:
			self.dispatch('on_open_failed', msg)
		if 'Invalid data found when processing input' in msg:
			self.dispatch('on_open_failed', msg)
		if 'Http error 403' in msg:
			self.dispatch('on_open_failed', msg)
		if 'End of file' in msg:
			self.dispatch('on_open_failed', msg)
		if 'I/O error' in msg:
			self.dispatch('on_open_failed', msg)
		if 'Server returned 4' in msg:
			self.dispatch('on_open_failed', msg)
		if 'Server returned 5' in msg:
			self.dispatch('on_open_failed', msg)
			
		msg = msg.strip()
		if msg:
			logger_func[level]('yffpyplayer: {}'.format(msg))

	def get_video_size(self):
		if hasattr(self._video,'_ffplayer'):
			try:
				return self._video._ffplayer.get_frame()[0][0].get_size()
			except:
				return None
		else:
			Logger.error('NewVideo: _video has not _ffplayer, do nothong')
		return None
		
	def audioswitch(self,btn=None):
		if hasattr(self._video,'_ffplayer'):
			x = self._video._ffplayer.request_channel('audio')
		else:
			Logger.error('NewVideo: _video has not _ffplayer, do nothong')

	def on_video_loaded(self, *args):
		if self.load_status == 'start':
			self.load_status = 'success'
			t = time.time()
			self.dispatch('on_load_success',
							{
								'source':self.source,
								'time':int((t - self.start_time)*100)
							})


	def on_state(self,*args):
		super().on_state(*args)
		if self.state == 'play':
			Window.allow_screensaver = False
		else:
			Window.allow_screensaver = True

	# fix bug for android
	# ValueError: None is not allowed for NewVideo.duration 
	def _on_video_frame(self, *largs):
		video = self._video
		if not video:
			return
		if video.duration is None:
			return
		if video.position is None:
			return
		super()._on_video_frame(*largs)

	#
	# try to fix up the error
	def unload(self,try_cnt=0):
		"""
		File "C:\Python37\lib\threading.py", line 1039, in join
		RuntimeError: cannot join thread before it is started
		"""
		try:
			super().unload()
			return
		except Exception as e:
			if try_cnt > 3:
				print('video.unload() tried 3 time,still error....',e)
				raise e
			time.sleep(0.1)
			return self.unload(try_cnt+1)



import sys
import os
from kivy.logger import Logger
from kivy.app import App
from kivy.properties import StringProperty
from kivy.core.audio.audio_ffpyplayer import SoundFFPy
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.progressbar import ProgressBar
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from kivy.clock import Clock
from kivy.uix.filechooser import FileChooserListView
from kivy.properties import ObjectProperty, StringProperty, BooleanProperty, \
    NumericProperty, DictProperty, OptionProperty

class APlayer(BoxLayout):
	source = StringProperty('')
	def __init__(self,afile=None,loop=False,
			can_openfile=False,
			can_changevolume=False,
			can_cut = False,
			can_replay = False,
			can_move=False):
		super().__init__(orientation='vertical')
		self.ap = None
		self.end_flag = False
		
	def on_source(self,o,s):
		if self.ap:
			self.ap.stop()
		self.ap = SoundFFPy(source=self.source)
		if not self.ap:
			return
		self.play()

	def __del__(self):
		self.end_flag=True
		self.stop()

	def stop(self):
		if not self.ap:
			return
		self.ap.stop()
		self.ap = None

	def play(self):
		if not self.ap:
			return 
		self.ap.bind(on_stop=self.end_play)
		self.ap.play()
		
	def end_play(self,obj):
		self.ap = None
		
if __name__ == '__main__':
	class MyApp(App):
		def build(self):
			af = None
			if len(sys.argv) > 1:
				af = sys.argv[1:]
			return APlayer()
	MyApp().run()



import traceback
import math

from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.clock import Clock
from kivy.app import App
from functools import partial
from appPublic.dictObject import DictObject
from appPublic.jsonConfig import getConfig
from .baseWidget import Text, HTTPDataHandler
from .utils import CSize, absurl, alert
from .form import StrSearchForm
from .dataloader import HttpDataLoader
from .dataloader import ListDataLoader
from .dataloader import RegisterFunctionDataLoader

class PagingButton(Button):
	def __init__(self, **kw):
		super().__init__(**kw)
		self.size_hint = (None,None)
		self.size = CSize(2,1.8)
		self.font_size = CSize(1)

"""
{
	dataurl
	params
	method
	locater
	filter
}

PageLoader load data in a page size once.
it fires two type of event
'on_newbegin':fire when start a new parameters loading series
'on_pageloaded':fire when a page data loads success
"""

class PageLoader(EventDispatcher):
	def __init__(self,target=None, **options):
		self.loading = False
		self.target = target
		self.options = options
		self.filter = None
		if options.get('filter'):
			self.filter = StrSearchForm(**options['filter'])
			self.filter.bind(on_submit=self.do_search)
		self.params = options.get('params',{})
		self.method = options.get('method','GET')
		self.url = options.get('dataurl')
		self.rfname = options.get('rfname')
		self.data = options.get('data')
		self.page_rows = options.get('page_rows',60)
		self.total_cnt = 0
		self.total_page = 0
		self.curpage = 0
		self.dir = 'down'
		self.register_event_type('on_newbegin')
		self.register_event_type('on_pageloaded')
		self.newbegin = True
		if self.url:
			self.loader = HttpDataLoader(self)
		elif self.rfname:
			self.loader = RegisterFunctionDataLoader(self)
		elif self.data:
			self.loader = ListDataLoader(self)
		else:
			raise Exception('need a url or rfname or data')
		self.loader.bind(on_success=self.show_page)
		self.loader.bind(on_error=self.onerror)
	
	def on_newbegin(self):
		pass

	def on_pageloaded(self,d):
		"""
		d={
			page:
			data:[]
			dir:'up' or 'down'
		}
		"""

	def do_search(self,o,params):
		self.newbegin = True
		self.total_cnt = 0
		self.total_page = 0
		self.curpage = 0
		self.dir = 'down'
		self.dispatch('on_newbegin')
		self.params.update(params)
		self.loadPage(1)

	def calculateTotalPage(self):
		self.total_page = math.floor(self.total_cnt / self.page_rows)
		if self.total_cnt % self.page_rows != 0:
			self.total_page += 1

	def setPageRows(self,row_cnt):
		if row_cnt == 0:
			return 
		self.page_rows = row_cnt
		if self.total_cnt != 0:
			self.calculateTotalPage()
	
	def show_page(self,o,d):
		p = (self.curpage - 1) * self.page_rows + 1
		for r in d['rows']:
			r['__posInSet__'] = p
			p += 1
		self.total_cnt = d['total']
		self.calculateTotalPage()
		d = {
			"page":self.curpage,
			"dir":self.dir,
			"data":d['rows']
		}
		self.dispatch('on_pageloaded',d)
		self.loading = False
		
	def onerror(self,o,e):
		traceback.print_exc()
		alert(str(e),title='alert')
		self.loading = False

	def loadNextPage(self):
		if self.loading:
			print('is loading, return')
			return

		if self.total_page > 0 and self.curpage >=self.total_page:
			return
		p = self.curpage + 1
		self.loadPage(p)

	def loadPreviousPage(self):
		if self.loading:
			print('is loading, return')
			return

		if self.curpage <= 1:
			return
		p = self.curpage - 1
		self.loadPage(p)

	def loadPage(self,p):
		if p == self.curpage:
			return

		if self.loading:
			print('is loading, return')
			return

		self.loading = True
		print(f'loading {p} page, {self.curpage}')
		if self.curpage > p:
			self.dir = 'up'
		else:
			self.dir = 'down'
		self.curpage = p
		self.loader.load()
	
"""
{
	adder,
	remover
	target,
	locater,
	dataurl,
	params,
	method,
	filter
}
events:
'on_deletepage': erase

"""
class RelatedLoader(PageLoader):
	def __init__(self, **options):
		super().__init__(**options)
		self.objectPages = {}
		self.totalObj = 0
		self.MaxbufferPage = 3
		self.locater = 1/self.MaxbufferPage
		if self.filter:
			self.widget = self.filter
		else:
			self.widget = None
		self.register_event_type('on_deletepage')

	def do_search(self,o,params):
		self.objectPages = {}
		self.totalObj = 0
		self.MaxbufferPage = 3
		self.locater = 1/self.MaxbufferPage
		super().do_search(o, params)
		
	def getLocater(self):
		if self.newbegin:
			self.newbegin = False
			return 1

		x = 1 / self.MaxbufferPage
		if self.dir != 'down':
			x = 1 - x
		return x

	def on_deletepage(self,d):
		pass

	def setPageRows(self,row_cnt):
		if self.total_cnt != 0:
			self.calculateTotalPage()
	
	def doBufferMaintain(self):
		siz = len(self.objectPages.keys())
		if siz >= self.MaxbufferPage:
			if self.dir == 'up':
				p = max(self.objectPages.keys())
			else:
				p = min(self.objectPages.keys())
			self.deleteBuffer(p)

	def deleteBuffer(self,page):
		d = self.objectPages[page]
		self.dispatch('on_deletepage',d)
		self.totalObj -= len(self.objectPages[page])
		del self.objectPages[page]
	
	def bufferObjects(self,page,objects):
		self.objectPages[page] = objects


	def show_page(self,o,data):
		if self.objectPages.get(self.curpage):
			self.deleteBuffer(self.curpage)
		else:
			self.doBufferMaintain()
		self.totalObj += len(data['rows'])
		super().show_page(o,data)
	
	def loadPreviousPage(self):
		pages = self.objectPages.keys()
		if len(pages) < 1:
			return

		page = min(pages)
		if page <= 1:
			return

		page -= 1
		self.loadPage(page)
	
	def loadNextPage(self):
		pages = self.objectPages.keys()
		if len(pages) == 0:
			return
		page = max(pages)
		if page>=self.total_page:
			return
		page += 1
		self.loadPage(page)

"""
{
	adder,
	clearer
	target,
	dataurl
	params,
	method
}
"""
class Paging(PageLoader):
	def __init__(self,**options):
		PageLoader.__init__(self,**options)
		self.target = options.get('target')
		self.init()

	def init(self):
		kwargs = {}
		kwargs['size_hint_y'] = None
		kwargs['height'] = CSize(2)
		kwargs['orientation'] = 'horizontal'
		kwargs['spacing'] = CSize(1)
		self.widget = BoxLayout(**kwargs)
		self.b_f = PagingButton(text="|<")
		self.b_p = PagingButton(text="<")
		self.b_n = PagingButton(text=">")
		self.b_l = PagingButton(text=">|")
		self.b_f.bind(on_press=self.loadFirstPage)
		self.b_p.bind(on_press=self.loadPreviousPage)
		self.b_n.bind(on_press=self.loadNextPage)
		self.b_l.bind(on_press=self.loadLastPage)
		self.widget.add_widget(self.b_f)
		self.widget.add_widget(self.b_p)
		self.widget.add_widget(self.b_n)
		self.widget.add_widget(self.b_l)
		if self.filter:
			self.widget.add_widget(self.filter)

	def loadFirstPage(self,o=None):
		if self.curpage == 1:
			return
		self.loadPage(1)
	
	def loadPreviousPage(self,o=None):
		if self.curpage < 2:
			return
		self.loadPage(self.curpage-1)
	
	def loadNextPage(self,o=None):
		if self.curpage >= self.total_page:
			return
		self.loadPage(self.curpage+1)
	
	def loadLastPage(self,o=None):
		if self.curpage >= self.total_page:
			return
		self.loadPage(self.total_page)




import time
try:
    import ujson as json
except:
	import json
from functools import partial

from kivy.logger import Logger
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.label import Label
from kivy.graphics import Fbo
from kivy.uix.button import ButtonBehavior
from kivy.clock import Clock
from kivy.properties import BooleanProperty, StringProperty, \
		OptionProperty, DictProperty, NumericProperty, ListProperty
from kivy.properties import ListProperty
from kivy.graphics import Color, Rectangle
from kivy.app import App
from kivy.factory import Factory

from appPublic.dictObject import DictObject
from appPublic.uniqueID import getID
from appPublic.myTE import string_template_render

from .utils import *
from .baseWidget import Text, HBox, VBox
from .scrollpanel import ScrollPanel
from .paging import Paging, RelatedLoader
from .ready import WidgetReady
from .toolbar import Toolbar
from .bgcolorbehavior import BGColorBehavior
from .widget_css import WidgetCSS
from .uitype.factory import UiFactory

class BLabel(ButtonBehavior, Text):
	def __init__(self, **kw):
		ButtonBehavior.__init__(self)
		Text.__init__(self,**kw)
		self.csscls = 'dummy'
		
class Cell(ButtonBehavior, WidgetCSS, BoxLayout):
	colume_name = StringProperty(None)
	cell_type = OptionProperty('data', \
			options=['data', 'header'])
	def __init__(self,row,desc, **kw):
		"""
		desc:{
			width,
			datatype:fff
			value:
			format:
			on_press:
		}
		"""
		self.desc = desc
		self.row = row
		super().__init__(size_hint=(None,None),
							width = self.desc['width'],
							height = self.row.part.datagrid.rowHeight()
		)
		self.csscls=self.row.part.datagrid.row_normal_css
		if self.row.header:
			self.csscls=self.row.part.datagrid.header_css
		if desc['header']:
			bl = Text(i18n=True, text=str(desc['value']),
				font_size=CSize(1),wrap=True,
				halign='left', valign='middle'
			)
			self.cell_type = 'header'
		else:
			self.cell_type = 'data'
			bl = UiFactory.build_view_widget(desc,self.row.row_data) 
		self.colume_name = desc['name']
		if bl:
			self.add_widget(bl)

class Row(BoxLayout):
	def __init__(self,part, rowdesc,header=False,data=None, **kw):
		"""
		rowdesc=[
			{
				width,
				name
				value
				on_press
			}

		"""
		self.part = part
		self.header = header
		self.row_data = data
		self.row_id = None
		self.linewidth = self.part.datagrid.linewidth
		self.rowdesc = rowdesc
		opts = kw.copy()
		opts.update({
			"spacing":self.linewidth,
			"orientation":"horizontal"
		})
		super(Row, self).__init__(**opts)
		self.height = self.part.datagrid.rowHeight()
		self.init(0)

	def on_row_press(self, *args):
		pass

	def init(self,t):
		w = 0
		h = 0
		for c in self.rowdesc:
			c['header'] = self.header
			cell = Cell(self,c)
			self.add_widget(cell)
			cell.bind(on_press=self.part.datagrid.cell_pressed)
			w += cell.width
		self.size_hint = None,None
		self.width = w + self.linewidth * (len(self.rowdesc)+1)

	def unselected(self):
		self.select(False)

	def selected(self):
		self.select(True)

	def select(self, flag):
		for c in self.children:
			if flag:
				c.csscls = self.part.datagrid.row_selected_css
			else:
				c.csscls = self.part.datagrid.row_normal_css

class Header(WidgetReady, ScrollPanel):
	def __init__(self,part,**kw):
		SUPER(Header, self, kw)
		self.part = part
		self.init(1)
		self.bind(on_scroll_stop=self.part.datagrid.scrollstop)
		if self.part.freeze_flag:
			self.bar_width = 0
		self.bar_width = 0

	def init(self,t):
		rd = [ f.copy() for f in self.part.rowdesc ]
		[ f.update({'value':self.part.fields[i].get('label', \
							self.part.fields[i].get('name'))}) \
							for i,f in enumerate(rd) ]
		self.header = Row(self.part,rd,header=True)
		self.add_widget(self.header)
		self.height = self.header.height

class Body(WidgetReady, ScrollPanel):
	def __init__(self,part,**kw):
		self.part = part
		SUPER(Body, self, kw)
		self._inner.spacing = self.part.datagrid.linewidth
		self.size_hint=(1,1)
		self.idRow = {}
		self.bind(on_scroll_stop=self.part.datagrid.scrollstop)
		if self.part.freeze_flag:
			self.bar_width = 0

	def addRow(self,id, data,index=0):
		rd = [ f.copy() for f in self.part.rowdesc ]
		[ f.update({'value':data.get(f['name'])}) for f in rd ]
		row = Row(self.part,rd,data=data)
		row.row_id = id
		self.add_widget(row,index=index)
		self.idRow[id] = row
		return row
	
	def clearRows(self):
		self.idRow = {}
		self.clear_widgets()

	def delRowById(self,id):
		row = self.idRow.get(id)
		if row:
			self.remove_widget(row)
		if self.idRow.get(id):
			del self.idRow[id]

	def getRowData(self,rowid):
		return self.idRow[rowid].row_data

	def getRowHeight(self):
		return self.part.datagrid.rowHeight()

	def get_row_by_id(self, rowid):
		return self.idRow.get(rowid)

class DataGridPart(WidgetReady, BoxLayout):
	def __init__(self,dg, freeze_flag, fields):
		self.datagrid = dg
		self.fields = fields
		self.freeze_flag = freeze_flag
		self.fields_width = None
		BoxLayout.__init__(self, orientation='vertical')
		WidgetReady.__init__(self)
		self.init(0)

	def setWidth(self):
		if self.freeze_flag:
			self.size_hint_x = None
			self.width = self.getFieldsWidth()

	def getFieldsWidth(self):
		if not self.fields_width:
			width = 0
			for f in self.rowdesc:
				width += f['width']
			self.fields_width =  width
		return self.fields_width

	def init(self,t):
		rd = []
		for f in self.fields:
			r = f.copy()
			r['width'] = CSize(f.get('width',10))
			rd.append(r)
		self.rowdesc = rd
		self.setWidth()
		kw = {}
		if self.freeze_flag:
			kw['width'] = self.fields_width
			kw['size_hint'] = (None,None)
		else:
			kw['size_hint'] = (1,None)
		kw['height'] = self.datagrid.rowHeight()
			
		self.header = None
		if not self.datagrid.noheader:
			self.header = Header(self,**kw)
			self.add_widget(self.header)
		inner = {
			"widgettype":"VBox",
			"options":{
			}
		}
		self.body = Body(self)
		self.add_widget(self.body)
		if not self.freeze_flag:
			self.body.bind(pos=self.datagrid.bodyOnSize,
							size=self.datagrid.bodyOnSize)
		
	def clearRows(self):
		return self.body.clearRows()

	def addRow(self,id, data):
		return self.body.addRow(id, data)

	def on_size(self, o, s=None):
		if not hasattr(self, 'body'):
			return
		if hasattr(self, '_toolbar'):
			if self._toolbar is not None:
				self._toolbar.width = self.width
		self.body.size_hint_y = None
		if self.header:
			self.body.height = self.height - self.header.height
		else:
			self.body.height = self.height


class DataGrid(VBox):
	"""
	DataGrid data format:
	{
		"widgettype":"DataGrid",
		"dataloader":{
			"page":1,
			"rows":60,
			"dataurl":"eeee",
			"params":{
			}
		},
		"tailer":{
			"options":{
			}
			"info":[
				"total_cnt",
				"total_page",
				"page_rows",
				"curpage"
			],
			"others":{
			}
		},
		"row_height": CSize,
		"header_css":"default",
		"body_css":"default",
		"spacing":1,
		"fields":[
			{
				"name":"field1",
				"label":"Field1",
				"datatype":"str",
				"uitype":"code",
				"value_field":,
				"text_field":
			},
			{
				"name":"field2",
				"label":"Field2",
				"viewer":{
					block dic
				}
			}
			...
		]
		"binds":[
		]
	}
	"""
	row_selected = BooleanProperty(False)
	row_normal_css = StringProperty('default')
	row_selected_css = StringProperty('default')
	header_css = StringProperty('default')
	body_css = StringProperty('default')
	row_height = NumericProperty(2)
	noheader = BooleanProperty(False)
	linewidth = NumericProperty(1)
	toolbar = DictProperty(None)
	dataloader = DictProperty(None)
	fields = ListProperty(None)
	tailer = DictProperty(None)
	def __init__(self,**options):
		self.select_rowid = None
		self.rowheight = None
		self.on_sizeTask = None
		self.selected_rowid = None
		self.show_rows = 0
		self._toolbar = None
		self.freeze_part = None
		self.normal_part = None
		SUPER(DataGrid, self, options)
		self.freeze_fields = self.getPartFields(freeze_flag=True)
		self.normal_fields = self.getPartFields(freeze_flag=False)
		if not self.dataloader:
			raise Exception('DataGrid need a DataLoader')
		self._dataloader = RelatedLoader(target=self, **self.dataloader)
		self._dataloader.bind(on_deletepage=self.delete_page)
		self._dataloader.bind(on_pageloaded=self.add_page)
		self._dataloader.bind(on_pageloaded=self.update_tailer_info)
		self._dataloader.bind(on_newbegin=self.clearRows)
		self.register_event_type('on_selected')
		self.register_event_type('on_rowpress')
		self.register_event_type('on_cellpress')
		self.register_event_type('on_headerpress')
		self.createDataGridPart()
		self.createToolbar()
		if self._toolbar:
			self.add_widget(self._toolbar)
		
		b = BoxLayout(orientation='horizontal')
		if self.freeze_part:
			b.add_widget(self.freeze_part)
		if self.normal_part:
			b.add_widget(self.normal_part)
		self.add_widget(b)
		if self.tailer:
			self.tailer_widgets = {}
			self.build_tailer(self.tailer)

	def on_rowpress(self, *args):
		print('on_rowpress fire, args=', args)

	def on_cellpress(self, *args):
		print('on_cesspress fire, args=', args)

	def on_headerpress(self, *args):
		print('on_headerpress fire, args=', args)

	def cell_pressed(self, o):
		if o.cell_type == 'header':
			self.dispatch('on_headerpress', o.colume_name)
			return
		row = o.row
		if self.selected_rowid:
			self.unselect_row(self.selected_rowid)
		
		self.selected_rowid = row.row_id
		self.select_row(row.row_id)
		self.dispatch('on_cellpress', o)
		self.dispatch('on_rowpress', row)
		self.dispatch('on_selected', row)

	def unselect_row(self, row_id):
		if self.freeze_part:
			row = self.freeze_part.body.get_row_by_id(row_id)
			if not row:
				return
			row.unselected()
		row = self.normal_part.body.get_row_by_id(row_id)
		if not row:
			return
		row.unselected()

	def select_row(self, row_id):
		if self.freeze_part:
			row = self.freeze_part.body.get_row_by_id(row_id)
			row.selected()
		row = self.normal_part.body.get_row_by_id(row_id)
		row.selected()
		
	def on_ready(self, *args):
		self.loadData()

	def build_tailer(self, tailer_desc):
		kw = tailer_desc.get('options', {})
		kw.update({
			'size_hint_y':None,
			'height':self.rowheight
		})
		w = HBox(**kw)
		self.add_widget(w)
		self.show_infos(w, tailer_desc.get('info'))
		if tailer_desc.get('others'):
			w1 = self.build_tailer_other(tailer_desc.get('others'))
			if w1:
				w.add_widget(w1)
	
	def update_tailer_info(self, *args):
		if not hasattr(self, 'tailer_widgets'):
			return
		for n,w in self.tailer_widgets.items():
			w.text = self.loader_info(n)

	def show_infos(self, tailer_widget, info_names):
		for n in info_names:
			desc = {
				"widgettype":"Text",
				"options":{
					"text":n,
					"i18n":True,
				}
			}
			w = Factory.Blocks().widgetBuild(desc)
			tailer_widget.add_widget(w)
			tailer_widget.add_widget(Label(text=':'))
			self.tailer_widgets[n] = Label(text=self.loader_info(n))
			tailer_widget.add_widget(self.tailer_widgets[n])

	def build_tailer_others(desc):
		return Factory.Blocks().widgetBuild(desc)

	def loader_info(self, n):
		if hasattr(self._dataloader, n):
			txt=getattr(self._dataloader, n, 0)
			if txt is None:
				txt = '0'
			txt = str(txt)
			return txt

	def locater(self,pos):
		self.normal_part.body.scroll_y = pos
		if self.freeze_part:
			self.freeze_part.body.scroll_y = pos

	def scrollstop(self,o,v=None):
		if not self.noheader and o == self.normal_part.header:
			self.normal_part.body.scroll_x = o.scroll_x
			return
		if o == self.normal_part.body:
			if not self.noheader:
				self.normal_part.header.scroll_x = o.scroll_x
			if self.freeze_part:
				self.freeze_part.body.scroll_y = o.scroll_y
		if self.freeze_part and o == self.freeze_part.body:
			self.normal_part.body.scroll_y = o.scroll_y

		if o.scroll_y <= 0.01:
			self._dataloader.loadNextPage()
		if o.scroll_y >= 0.99:
			self._dataloader.loadPreviousPage()

	def getValue(self):
		if not self.select_rowid:
			return None
		return self._getRowData(self.select_rowid)
	
	def _getRowData(self, rowid):
		d = {}
		if self.freeze_part:
			d.update(self.freeze_part.body.getRowData(rowid))
		d.update(self.normal_part.body.getRowData(rowid))
		return DictObject(**d)

	def bodyOnSize(self,o,s):
		if self.on_sizeTask is not None:
			self.on_sizeTask.cancel()
		self.on_sizeTask = Clock.schedule_once(self.calculateShowRows,0.3)

	def rowHeight(self):
		if not self.rowheight:
			self.rowheight = CSize(self.row_height)
		return self.rowheight
	
	def calculateShowRows(self,t):
		self.getShowRows()
		self._dataloader.setPageRows(self.show_rows * 2)

	def getShowRows(self):
		if self.show_rows == 0:
			return 60
		self.show_rows = int(self.rowHeight() / self.normal_part.body.height)
		return self.show_rows * 2

	def clearRows(self, *args):
		if self.freeze_part:
			self.freeze_part.body.clearRows()
		self.normal_part.body.clearRows()

	def add_page(self,o,data):
		dir = data['dir']
		if not self.show_rows:	
			self.getShowRows()

		ids = []
		recs = data['data']
		idx = 0
		if dir == 'up':
			recs.reverse()
			idx = -1
		recs1 = recs[:self.show_rows]
		recs2 = recs[self.show_rows:]
		self._fbo = Fbo(size=self.size)
		with self._fbo:
			self._background_color = Color(0,0,0,1)
			self._background_rect = Rectangle(size=self.size)
		for r in recs1:
			id = self.addRow(r,index=idx)
			ids.append(id)
		with self.canvas:
			self._fbo_rect = Rectangle(size=self.size,
								texture=self._fbo.texture)

		data['idx'] = idx
		data['ids'] = ids
		data['data'] = recs2
		f = partial(self.add_page_delay,data)
		Clock.schedule_once(f, 0)

	def add_page_delay(self, data, *args):
		recs = data['data']
		page = data['page']
		idx = data['idx']
		ids = data['ids']
		self._fbo = Fbo(size=self.size)
		with self._fbo:
			self._background_color = Color(0,0,0,1)
			self._background_rect = Rectangle(size=self.size)
		for r in recs:
			id = self.addRow(r,index=idx)
			ids.append(id)
		with self.canvas:
			self._fbo_rect = Rectangle(size=self.size,
								texture=self._fbo.texture)
		self._dataloader.bufferObjects(page,ids)
		x = self._dataloader.getLocater()
		self.locater(x)

	def delete_page(self,o,data):
		for id in data:
			self.delRow(id)

	def addRow(self,data, **kw):
		id = getID()
		f_row = None
		if self.freeze_part:
			self.freeze_part.body.addRow(id, data, **kw)
		self.normal_part.body.addRow(id, data, **kw)
		return id

	def delRow(self,id,**kw):
		if self.freeze_part:
			self.freeze_part.body.delRowById(id)
		self.normal_part.body.delRowById(id)

	def createToolbar(self):
		if self.toolbar:
			self._toolbar = Toolbar(**self.toolbar)

	def on_selected(self,row):
		print("DataGrid():on_selected fire")

	def loadData(self,*args, **kwargs):
		kwargs['page'] = 1
		self.selected_rowid = None
		self._dataloader.do_search(None,kwargs)

	def createDataGridPart(self):
		self.freeze_part = None
		self.normal_part = None
		if self.freeze_fields:
			self.freeze_part = DataGridPart(self,True, self.freeze_fields)
		if self.normal_fields:
			self.normal_part = DataGridPart(self, False, self.normal_fields)

	def getPartFields(self,freeze_flag:bool=False) -> list:
		fs = []
		for f in self.fields:
			if freeze_flag:
				if f.get('freeze',False):
					fs.append(f)
			else:
				if not f.get('freeze',False):
					fs.append(f)
		return fs
	
	def get_selected_data(self):
		if not self.selected_rowid:
			return {}
		data = {}
		row_id = self.selected_rowid
		if self.freeze_part:
			row = self.freeze_part.body.get_row_by_id(row_id)
			data.update(row.row_data)
		row = self.normal_part.body.get_row_by_id(row_id)
		data.update(row.row_data)
		return data



import codecs
from .dataloader import DataLoader

class FileDataLoader(DataLoader):
	def loadData(self,filename):
		try:
			with codecs.open(filename,'r','utf8') as f:
				self.dataLoaded(text)
				return f.read()
		except Exception as e:
			self.loadError(e)





from kivy.uix.label import Label
from kivyblocks.widgetExt.inputext import StrInput
from ..utils import *

view_widget_builders = {
}

input_widget_builders = {
}

def get_value(desc, rec=None):
	if rec:
		return rec.get(desc['name'])
	return desc.get('value', desc.get('defaultvalue', ''))

class UiFactory:
	@classmethod
	def register(self, ut, inpf, viewf):
		view_widget_builders.update({ut:viewf})
		input_widget_builders.update({ut:inpf})

	@classmethod
	def build_view_widget(self, desc, rec=None):
		ut = desc.get('uitype', 'str')
		f = view_widget_builders.get(ut)
		if f is None:
			print(desc, 'view builder not defined')
			return None
		return f(desc, rec=rec)

	@classmethod
	def build_input_widget(self, desc, rec=None):
		ut = desc.get('uitype', 'str')
		f = input_widget_builders.get(ut)
		if f is None:
			print(desc, 'input builder not defined')
			return None
		return f(desc, rec=rec)

class StringView(Label):
	@classmethod 
	def builder(self, desc, rec):
		txt = get_value(desc, rec=rec)
		return StringView(text=str(txt) if txt else '',
						font_size=CSize(1),
						halign='left',
						valign='middle')

	def __init__(self, **kw):
		super(StringView, self).__init__(**kw)
		self.wrap = True
		self.halign = 'left'
		self.valign = 'middle'

	def getValue(self):
		return self.text

	def setValue(self, v):
		self.text = str(v)

class StringInput(StrInput):
	@classmethod
	def builder(self, desc, rec):
		v = get_value(desc, rec=rec)
		kw = desc.get('uiparams', {})
		kw['multiline'] = False
		w = StringInput(**kw)
		if v:
			w.setValue(v)
		return w

UiFactory.register('str', StringInput.builder, StringView.builder)




import time
from kivy.logger import Logger

class SwipeBehavior(object):
	def __init__(self, **kwargs):
		object.__init__(self)
		self.swipeenable = True
		self.register_event_type('on_context_menu')
		self.register_event_type('on_swipe_left')
		self.register_event_type('on_swipe_right')
		self.register_event_type('on_swipe_up')
		self.register_event_type('on_swipe_down')
		self.sb_start_point = None
		self.sb_end_point = None
		self.sb_start_time = None
		self.sb_end_time = None
		self.threshold=20
		self.threshold_time = 0.6
		self.bind(on_touch_down=self.on_touchdown)
		self.bind(on_touch_move=self.on_touchmove)
		self.bind(on_touch_up=self.on_touchup)

	def disable_swipe(self):
		self.swipeenable = False

	def enable_swipe(self):
		self.swipeenable = True

	def on_context_menu(self):
		pass
		# Logger.info('SwipeBehavior:on_context_menu fired')

	def on_swipe_left(self):
		pass
		# Logger.info('SwipeBehavior:on_swipe_left fired')

	def on_swipe_right(self):
		pass
		# Logger.info('SwipeBehavior:on_swipe_right fired')

	def on_swipe_up(self):
		pass
		# Logger.info('SwipeBehavior:on_swipe_up fired')

	def on_swipe_down(self):
		pass
		# Logger.info('SwipeBehavior:on_swipe_down fired')

	def on_touchdown(self,o, touch):
		if touch.is_mouse_scrolling:
			# Logger.info('SwipeBehavior:is_mouse_scrolling')
			return False
		if not self.collide_point(touch.x, touch.y):
			return False

		# Logger.info('SwipeBehavior:touch_down fired')
		if self.swipeenable:
			# Logger.info('SwipeBehavior:touch_down fired')
			self.sb_start_point = touch.pos
			self.sb_start_time = time.time()

	def on_touchmove(self,o,touch):
		if self.collide_point(*touch.pos) and self.swipeenable:
			# Logger.info('SwipeBehavior:touch_move fired')
			if self.sb_start_point is None:
				self.sb_start_point = touch.pos
			else:
				self.sb_end_point = touch.pos

	def on_touchup(self,o,touch):
		ret = False
		if self.collide_point(*touch.pos) and self.swipeenable:
			# Logger.info('SwipeBehavior:touch_up fired')
			self.sb_end_point = touch.pos
			self.sb_end_time = time.time()
			self.check_context_menu()
			ret = True
		self.check_swipe()
		self.sb_start_point = None
		self.sb_end_point = None
		self.sb_start_time = None
		self.sb_end_time = None

	def check_context_menu(self):
		if not self.sb_start_time:
			return
		if not self.sb_end_time:
			return
		period_time = self.sb_end_time - self.sb_start_time
		# Logger.info('SwipeBehavior:period_time=%f,threshold_time=%f',
		#				period_time, self.threshold_time)
		if period_time >= self.threshold_time:
			self.dispatch('on_context_menu')

	def check_swipe(self):
		if not self.sb_end_point:
			return

		if not self.sb_start_point:
			return

		if abs(self.sb_end_point[0] - self.sb_start_point[0]) > \
					abs(self.sb_end_point[1] - self.sb_start_point[1]):
			Logger.info('SwipeBehavior:check_swipe x>y')
			if self.sb_end_point[0] - self.sb_start_point[0] >= \
						self.threshold:
				self.dispatch('on_swipe_right')
			elif self.sb_start_point[0] - self.sb_end_point[0] >= \
						self.threshold:
				self.dispatch('on_swipe_left')
		else:
			# Logger.info('SwipeBehavior:check_swipe x<y')
			if self.sb_end_point[1] - self.sb_start_point[1] >= \
						self.threshold:
				self.dispatch('on_swipe_up')
			elif self.sb_start_point[1] - self.sb_end_point[1] >= \
						self.threshold:
				self.dispatch('on_swipe_down')



from kivy.utils import get_color_from_hex
#from kivy.app import App
#from kivy.uix.widget import Widget
#from kivy.graphics import Color, Rectangle, Ellipse
#from kivy.uix.boxlayout import BoxLayout
import math

def toArrayColor(color):
	if isinstance(color,str):
		color = get_color_from_hex(color)
		if len(color) == 3:
			color.append(1)
	return color
	
def distance(color1,color2):
	color1 = toArrayColor(color1)
	color2 = toArrayColor(color2)
	x = math.pow(abs(color2[0] - color1[0]),2) + \
		math.pow(abs(color2[1] - color1[1]),2) + \
		math.pow(abs(color2[2] - color1[2]),2)
	d = math.sqrt(x)
	return d

def reverseColor(color):
	center = [0.5,0.5,0.5]
	if color == [0.5,0.5,0.5]:
		return [1,1,1]
	d = distance(color,center)
	d1 = d + 0.5
	point= [0,0,0]
	for i in range(3):
		m = color[i] - center[i]
		m1 = abs(m) * d1 / d
		point[i] = color[i] - m1 if m < 0 else color[i] + m1
	return point

#DIVIDE
def divideColor(color1,color2,divide_cnt):
	color1 = toArrayColor(color1)
	color2 = toArrayColor(color2)
	return divide(color1,color2,divide_cnt)

def divide(point1,point2,divide_cnt):
	ps = [point1]
	dim = len(point1)
	for i in range(1,divide_cnt):
		p = []
		cnt = divide_cnt
		for j in range(dim):
			m = (i*point2[j] + (cnt - i) * point1[j]) / cnt
			p.append(m)
		ps.append(p)
	ps.append(point2)
	return ps



from kivy.utils import platform
from .binstateimage import BinStateImage
from .jsoncodeinput import JsonCodeInput
from .inputext import FloatInput,IntegerInput,StrInput,SelectInput, BoolInput, Password
from .scrollwidget import ScrollWidget
from .messager import Messager
__all__ = [
BinStateImage,
JsonCodeInput,
FloatInput,
Password,
BoolInput,
IntegerInput,
StrInput,
SelectInput,
ScrollWidget,
Messager,
]

if platform == 'android':
	print('***********************************8')
	from .phonebutton import PhoneButton
	from .androidwebview import AWebView
	__all__ = __all__ + [PhoneButton, AWebView]



from kivy.uix.scrollview import ScrollView
from kivy.effects.scroll import ScrollEffect
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.graphics import Color, Ellipse,Rectangle

class ScrollWidget(ScrollView):
	def __init__(self,**kw):
		super(ScrollWidget,self).__init__()
		self.effect_cls = ScrollEffect
		self.sized_widgets = []
		kw.update({
			'size_hint':(None,None),
			'orientation':'vertical'
		})
		self._inner = BoxLayout(**kw)
		self._inner.bind(
				minimum_height=self._inner.setter('height'))
		self._inner.bind(
				minimum_width=self._inner.setter('width'))
		super(ScrollWidget,self).add_widget(self._inner)
		self.bind(size=self.change_all_sized_widgets)

	def change_all_sized_widgets(self, *args):
		for w in self.sized_widgets:
			self.change_widget_width(w)
		self._inner.do_layout()

	def change_widget_width(self, w):
		w.size_hint_x = None
		w.width = self.width * w.org_size_hint_x

	def add_widget(self,widget,**kw):
		if widget.size_hint_x:
			widget.org_size_hint_x = widget.size_hint_x
			self.change_widget_width(widget)
			self.sized_widgets.append(widget)
		a = self._inner.add_widget(widget,**kw)
		return a

	def clear_widgets(self,**kw):
		self.sized_widgets = []
		a = self._inner.clear_widgets(**kw)

	def remove_widget(self,widget,**kw):
		self.sized_widgets = [i for i in self.sized_widgets if i != widget]
		a = self._inner.remove_widget(widget,**kw)
		return a

if __name__ == '__main__':
	from kivy.app import App
	from kivy.uix.label import Label
	from kivy.uix.button import Button
	import codecs
	
	class MyApp(App):
		def build(self):
			root = ScrollWidget(size=(400,400),
					pos_hint={'center_x': .5, 'center_y': .5}
			)
			with codecs.open(__file__,'r','utf-8') as f:
				txt = f.read()
				lines = txt.split('\n')
				for l in lines:
					root.add_widget(Label(text=l,color=(1,1,1,1),size_hint=(None,None),size=(1200,40)))
			return root

	MyApp().run()



# coding=utf-8

__all__ = ["Downloader"]

import logging
import traceback
from concurrent.futures import ThreadPoolExecutor, TimeoutError, as_completed
from os import environ, makedirs
from os.path import exists, join
from random import choice
from time import time

import requests
from kivy.clock import Clock
from kivy.logger import LOG_LEVELS, Logger

from kivyblocks.mapview.constants import CACHE_DIR

if "MAPVIEW_DEBUG_DOWNLOADER" in environ:
    Logger.setLevel(LOG_LEVELS['debug'])

# user agent is needed because since may 2019 OSM gives me a 429 or 403 server error
# I tried it with a simpler one (just Mozilla/5.0) this also gets rejected
USER_AGENT = 'Kivy-garden.mapview'


class Downloader:
    _instance = None
    MAX_WORKERS = 5
    CAP_TIME = 0.064  # 15 FPS

    @staticmethod
    def instance(cache_dir=None):
        if Downloader._instance is None:
            if not cache_dir:
                cache_dir = CACHE_DIR
            Downloader._instance = Downloader(cache_dir=cache_dir)
        return Downloader._instance

    def __init__(self, max_workers=None, cap_time=None, **kwargs):
        self.cache_dir = kwargs.get('cache_dir', CACHE_DIR)
        if max_workers is None:
            max_workers = Downloader.MAX_WORKERS
        if cap_time is None:
            cap_time = Downloader.CAP_TIME
        self.is_paused = False
        self.cap_time = cap_time
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self._futures = []
        Clock.schedule_interval(self._check_executor, 1 / 60.0)
        if not exists(self.cache_dir):
            makedirs(self.cache_dir)

    def submit(self, f, *args, **kwargs):
        future = self.executor.submit(f, *args, **kwargs)
        self._futures.append(future)

    def download_tile(self, tile):
        Logger.debug(
            "Downloader: queue(tile) zoom={} x={} y={}".format(
                tile.zoom, tile.tile_x, tile.tile_y
            )
        )
        future = self.executor.submit(self._load_tile, tile)
        self._futures.append(future)

    def download(self, url, callback, **kwargs):
        Logger.debug("Downloader: queue(url) {}".format(url))
        future = self.executor.submit(self._download_url, url, callback, kwargs)
        self._futures.append(future)

    def _download_url(self, url, callback, kwargs):
        Logger.debug("Downloader: download(url) {}".format(url))
        response = requests.get(url, **kwargs)
        response.raise_for_status()
        return callback, (url, response)

    def _load_tile(self, tile):
        if tile.state == "done":
            return
        cache_fn = tile.cache_fn
        if exists(cache_fn):
            Logger.debug("Downloader: use cache {}".format(cache_fn))
            return tile.set_source, (cache_fn,)
        tile_y = tile.map_source.get_row_count(tile.zoom) - tile.tile_y - 1
        uri = tile.map_source.url.format(
            z=tile.zoom, x=tile.tile_x, y=tile_y, s=choice(tile.map_source.subdomains)
        )
        Logger.debug("Downloader: download(tile) {}".format(uri))
        response = requests.get(uri, headers={'User-agent': USER_AGENT}, timeout=5)
        try:
            response.raise_for_status()
            data = response.content
            with open(cache_fn, "wb") as fd:
                fd.write(data)
            Logger.debug("Downloaded {} bytes: {}".format(len(data), uri))
            return tile.set_source, (cache_fn,)
        except Exception as e:
            print("Downloader error: {!r}".format(e))

    def _check_executor(self, dt):
        start = time()
        try:
            for future in as_completed(self._futures[:], 0):
                self._futures.remove(future)
                try:
                    result = future.result()
                except Exception:
                    traceback.print_exc()
                    # make an error tile?
                    continue
                if result is None:
                    continue
                callback, args = result
                callback(*args)

                # capped executor in time, in order to prevent too much
                # slowiness.
                # seems to works quite great with big zoom-in/out
                if time() - start > self.cap_time:
                    break
        except TimeoutError:
            pass



from kivy.properties import StringProperty, NumericProperty, ListProperty
from kivy.factory import Factory
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.uix.scrollview import ScrollView
from kivy.uix.label import Label
from kivy.uix.image import AsyncImage

from .baseWidget import ScrollWidget,  getDataHandler, VBox

from .utils import CSize
import re

SOURCE_DELIMITER = "\n```\n"
VIDEO_LEAD = "![@"
AUDIO_LEAD = "![#"
IMAGE_LEAD = "!["
LINK_LEAD = "["
REFER_SPLIT = "]("
REFER_END = ")"

class MarkdownParser(object):
	mdkeys = [
		SOURCE_DELIMITER,
		VIDEO_LEAD,
		AUDIO_LEAD,
		IMAGE_LEAD
	]
	def __init__(self, mdtext):
		self.mdtext = mdtext
		self.result = []
		self.text = ""
		self.mdkey_handler = {
			SOURCE_DELIMITER:self.handle_source_delimiter,
			VIDEO_LEAD:self.lead_video,
			AUDIO_LEAD:self.lead_audio,
			IMAGE_LEAD:self.lead_image,
		}

	def handle_source_delimiter(self):
		x = self.mdtext.split(SOURCE_DELIMITER,1)
		mk_p = MarkdownParser(x[0])
		d = {
			'source':x[0]
		}
		self.result.append(d) 
		if len(x) > 1:
			self.mdtext = x[1]

	def lead_video(self):
		print('lead_video() ..., mdtext=', self.mdtext)
		x = self.mdtext.split(REFER_SPLIT,1)
		if len(x)<2:
			print('lead_video() return here', REFER_SPLIT, '---', self.mdtext)
			self.text = f'{VIDEO_LEAD}'
			return
		title = x[0]
		y = x[1].split(REFER_END,1)
		if len(y) < 2:
			print('lead_video() return here', REFER_END)
			self.text = f'{VIDEO_LEAD}'
			return
		url = y[0]
		self.mdtext = y[1]
		d = {
			"video":{
				"title":title,
				"url":url
			}
		}
		self.result.append(d)

	def lead_audio(self):
		x = self.mdtext.split(REFER_SPLIT,1)
		if len(x)<2:
			self.text = f'{AUDIO_LEAD}'
			return
		title = x[0]
		y = x[1].split(REFER_END,1)
		if len(y) < 2:
			self.text = f'{AUDIO_LEAD}'
			return
		url = y[0]
		d = {
			"audio":{
				"title":"title",
				"url":url
			}
		}
		self.result.append(d)
		l = len(title) + 2 + len(url)
		self.mdtext = self.mdtext[l:]

	def lead_image(self):
		x = self.mdtext.split(REFER_SPLIT,1)
		if len(x)<2:
			self.text = f'{IMAGE_LEAD}'
			return
		title = x[0]
		y = x[1].split(REFER_END,1)
		if len(y) < 2:
			self.text = f'{IMAGE_LEAD}'
			return
		url = y[0]
		d = {
			"image":{
				"title":"title",
				"url":url
			}
		}
		self.result.append(d)
		self.mdtext = y[1]

	def check_key(self,t):
		for k in self.mdkeys:
			if t.startswith(k):
				return k
		return None

	def parse(self):
		"""
		parser parse mdtext, recognize bbtext, source, img, audio, video
		part text
		"""
		while len(self.mdtext) > 0:
			k = self.check_key(self.mdtext)
			if k is None:
				self.text = f'{self.text}{self.mdtext[0]}'
				self.mdtext = self.mdtext[1:]
			else:
				if len(self.text) > 0:
					self.result.append({'pure_md':self.text})
					self.text = ""
				self.mdtext = self.mdtext[len(k):]
				self.mdkey_handler[k]()
		if len(self.text) > 0:
			self.result.append({'pure_md':self.text})
			self.text = ''
		return self.result	

class MDImage(AsyncImage):
	parent_width=NumericProperty(None)
	image_size = ListProperty(None)	
	def __init__(self, **kw):
		super().__init__(**kw)
		self.image_size = None
		self.parent_width = None
		self.bind(texture=self.image_loaded)

	def image_loaded(self,o, *args):
		self.image_size = self.texture.size

	def resize(self, *args):
		ps = [0,0,0,0]
		if hasattr(self.parent, 'padding'):
			ps = self.parent.padding
		self.parent_width = self.parent.width - ps[0] - ps[2]
		if self.texture:
			if self.texture.size:
				self.image_size = self.texture.size
		Logger.info('MDImage:resize called, %d', self.parent_width)

	def on_image_size(self, *args):
		if self.parent_width:
			Logger.info('MDImage:on_iage_size %d, %s', self.parent_width,
						self.image_size)
			self.set_image_size()

	def on_parent_width(self, *args):
		if self.image_size:
			Logger.info('MDImage:on_oarent_width %d, %s', self.parent_width,
						self.image_size)
			self.set_image_size()

	def set_image_size(self):
		self.width = self.parent_width
		self.height = self.parent_width * self.image_size[1] \
							/ self.image_size[0]
		Logger.info('MDImage:set_image_size %d, %d', self.width,
					self.height)
	
class MarkdownBody(VBox):
	def __init__(self, md_obj=None, padding=[10,10,10,10], **kw):
		self.part_handlers = {
			"pure_md":self.build_pure_md,
			"source":self.build_source,
			"image":self.build_image,
			"video":self.build_video,
			"audio":self.build_audio
		}
		self.md_obj = md_obj
		super().__init__(**kw)
		self.padding=padding
		self.size_hint = None,None
		self.bind(parent=self.resize)
		self.resize()
	
	def show_mdtext(self, mdtext):
		mdp = MarkdownParser(mdtext)
		parts = mdp.parse()
		self.clear_widgets()
		for p in parts:
			for k,v in p.items():
				f = self.part_handlers.get(k)
				f(v)

	def resize(self, *args):
		Logger.info('MDBody:resize called')
		if self.parent:
			ps = [0,0,0,0]
			if hasattr(self.parent, 'padding'):
				ps = self.parent.padding
			h = 0
			for c in self.children:
				if hasattr(c, 'resize'):
					c.resize()
				h += c.height
			self.width = self.parent.width - ps[0] - ps[2]
			self.height = h
		else:
			Logger.info('resize:parent is null')

	def build_source(self,source_desc):
		w = MarkdownBody(md_obj=self.md_obj,
				csscls=self.md_obj.second_css, size_hint_y=None)
		w.show_mdtext(source_desc)
		self.add_widget(w)
		w.resize()

	def build_pure_md(self, mdtext):
		for l in mdtext.split('\n'):
			self.parse_line(l)

	def build_image(self,img_desc):
		w = MDImage(source=img_desc['url'],
					allow_stretch=True,
					size_hint_y=None,
					height=CSize(10),
					keep_ratio=True
			)
		
		self.add_widget(w)
		w.resize()

	def build_video(self, video_desc):
		w = Factory.NewVideo(source=video_desc['url'],
				keep_ratio=True,
				play=False,
				allow_stretch = True,
				size_hint_y=None
		)
		w.height=self.width * 10 / 16
		w.state = 'pause'
		def f1(x):
			x.state = 'play'
		def f2(x):
			x.state = 'stop'

		w.bind(on_enter_focus=f1)
		w.bind(on_leave_focus=f2)
		self.add_widget(w)
		w.resize()
		
	def build_audio(self, audio_desc):
		w = Factory.APlayer(source=audio_desc.url)
		w.bind(minimum_height=w.setter('height'))
		self.add_widget(w)

	def parse_title(self, txt, level):
		w = Factory.Blocks().widgetBuild({
			"widgettype":f"Title{level}",
			"options":{
				"text":txt,
				"size_hint_x":None,
				"width":self.width,
				"size_hint_y":None,
				"markup":True,
				"wrap":True,
				"halign":"left",
				"valign":"middle"
			}
		})
		if not w:
			return
		w1,h1 = w.get_wraped_size()
		clen = CSize(1)
		if h1 is None or h1 < clen:
			h1 = clen
		w.height = h1
		w.bind(on_ref_press=self.md_obj.open_new_md)
		self.add_widget(w)
		
	def parse_line(self, l):
		if l.startswith('###### '):
			t = self.mktext_bbtext(l[7:])
			return self.parse_title(t,6)
		if l.startswith('##### '):
			t = self.mktext2bbtext(l[6:])
			return self.parse_title(t,5)
		if l.startswith('#### '):
			t = self.mktext2bbtext(l[5:])
			return self.parse_title(t,4)
		if l.startswith('### '):
			t = self.mktext2bbtext(l[4:])
			return self.parse_title(t,3)
		if l.startswith('## '):
			t = self.mktext2bbtext(l[3:])
			return self.parse_title(t,2)
		if l.startswith('# '):
			t = self.mktext2bbtext(l[2:])
			return self.parse_title(t,1)
		t = self.mktext2bbtext(l)
		print('markdown.py, t=', t, 'len(t)=', len(t), 'type(t)=', type(t))
		w = Factory.Blocks().widgetBuild({
			"widgettype":"Text",
			"options":{
				"text":t,
				"wrap":True,
				"size_hint_x":None,
				"width":self.width,
				"markup":True,
				"valign":"middle",
				"halign":"left"
			}
		})
		if not w:
			return
		_,h = w.get_wraped_size()
		clen = CSize(1)
		if h is None or h < clen:
			h = clen
		w.height = h
		w.bind(on_ref_press=self.md_obj.open_new_md)
		self.add_widget(w)

	def mktext2bbtext(self,mdtext):
		"""
		convert markdown text to bbtag text kivy Label Widget recognized

		markdown syntax
		*XXX*		Italic
		**XXX**		Bold
		***XXX***	Bold + Italic
		> XXXX		reference
		>> XXXX		reference inside reference
		[-*] XXX	list
		[0-9]*[.] XXX	list
		***			split line
		---			split line
		___			split line
		[.*](url)	link
		![.*](url)	image
		[![.*](url)](url)		image link
		!v[.*](url)	video*		plan to extend
		!a[.*](url) audio*		plan to extend

		bb tag syntax
		[b]XXX[/b]	Bold
		[i]XXX[/i]	Italic
		[sub]XXX[/sub]
		"""
		mdtext = re.sub('\*\*\*(.*)\*\*\*', \
					lambda x: '[b][i]'+x.group(1)+'[/i][/b]', \
					mdtext)

		mdtext = re.sub('\*\*(.*)\*\*', \
					lambda x: '[b]'+x.group(1)+'[/b]', \
					mdtext)

		mdtext = re.sub('\*(.*)\*', \
					lambda x: '[i]'+x.group(1)+'[/i]', \
					mdtext)

		mdtext = re.sub('([^!]?)\[(.*)\]\((.*)\)', \
					lambda x:x.group(1) + '[ref='+x.group(3)+'][color=00ff00]'+ x.group(2)+'[/color][/ref]', \
					mdtext)

		return mdtext


class Markdown(ScrollView):
	"""
# Markdown
MArkdown widget using to render a markdown file and show it
description file format
{
	"widgettype":"Markdown",
	"options":{
		"source": the markdown file
		"first_css":,
		"second_css":
		other options
	}
}
	"""
	source = StringProperty(None)
	first_css = StringProperty("default")
	second_css = StringProperty("default")
	def __init__(self, **kw):
		super(Markdown, self).__init__(**kw)
		self.root_body = MarkdownBody(md_obj=self,
					csscls=self.first_css,
					size_hint_y=None
		)
		self.root_body.bind(minimum_height=self.root_body.setter('height'))
		self.add_widget(self.root_body)
		self.source = kw.get('source')
		if self.source:
			self.load_text()
		self.bind(size=self.root_body.resize)
		self.bind(source=self.load_text)

	def check_parent_window(self, *args):
		pw = self.get_parent_window()
		rw = self.get_root_window()
		print('MD: pw=%s,rw=%s', pw,rw)

	def stop_media(self, *args):
		self.root_body.stp_video()

	def load_text(self, *args):
		h = getDataHandler(self.source)
		h.bind(on_success=self.update)
		h.bind(on_error=self.show_error)
		h.handle()

	def show_error(self, o, e):
		print('load_text(), error=',e)

	def update(self, o, text):
		print('text=',text, type(text))
		if not text:
			return
		text = ''.join(text.split('\r'))
		self.root_body.show_mdtext(text)
		self.scroll_y = 1

	def open_new_md(self, o, value):
		print(value,'is it a link')
		self.source = value




from ..baseWidget import FloatInput
from ..utils import *
from .float import FloatView
from .factory import UiFactory, get_value

def amountstr(amount,dec):
	f = '%%.0%df' % dec
	s = f % amount
	parts = s.split('.')
	p_int = parts[0]
	subs = []
	if p_int[0] == '-':
		subs.append('-')
		p_int = p_int[1:]
	l = [i for i in p_int]
	l.reverse()
	k = []
	for i,c in enumerate(l):
		if i and (i) % 3 == 0:
			k.append(',')
		k.append(c)
	k.reverse()
	subs.append(''.join(k))
	if len(parts) > 1:
		subs.append(f'.{parts[1]}')
	return ''.join(subs)

class AmountView(FloatView):
	def on_value(self, *args):
		if self.value is None:
			self.text = self.nullstr
		else:
			self.text = self.amount_str()
	
	def amount_str(self):
		return amountstr(self.value, self.desc)

	def builder(self, desc, rec=None):
		v = get_value(desc, rec=rec)
		defaultvalue = desc.get('nullstr', '')
		return AmountView(value=v,
						dec=desc.get('dec', 2),
						nullstr=defaultvale,
						font_size=CSize(1), 
						halign='right', 
						valign='middle'
		)

class AmountInput(FloatInput):
	def input_filter(self,substring, from_undo=False):
		s = super(AmountInput,self).filter(substring)
		a = s.split('.')
		b = a[0]
		if len(b)>3:
			k = []
			while len(b)>3:
				x = b[-3:]
				k.insert(0,x)
				b = b[:-3]
			a[0] = ','.join(k)
		s = '.'.join(a)
		return '.'.join(a)

	def insert_text(self, substring, from_undo=False):
		x = StrInput.insert_text(self,s, from_undo=from_undo)
		y = amountstr(x, self.dec)
		return y

	def getValue(self):
		txt = ''.join(self.text.split(','))
		return float(txt)

	@classmethod
	def builder(self, desc, rec=None):
		v = get_value(desc, rec=rec)
		kw = desc.get('uiparams', {})
		w = AmountInput(**kw)
		if v is not None:
			w.setValue(v)
		return w

UiFactory.register('amount', AmountInput.builder, AmountView.builder)




# -*- coding: utf-8 -*-
from kivy.app import App
from kivy.core.text import LabelBase
from kivy.core.window import Window
from kivy.clock import Clock
from kivy.metrics import dp
from kivy.properties import OptionProperty, AliasProperty, ObjectProperty, \
    StringProperty, ListProperty, BooleanProperty
from kivy.uix.widget import Widget
from kivy.utils import get_color_from_hex
from kivy.atlas import Atlas
from kivyblocks.color_definitions import colors, text_colors
from kivyblocks.material_resources import FONTS, DEVICE_TYPE
from kivyblocks import images_path

for font in FONTS:
    LabelBase.register(**font)


class ThemeManager(Widget):
    primary_palette = OptionProperty(
        'Blue',
        options=['Pink', 'Blue', 'Indigo', 'BlueGrey', 'Brown',
                 'LightBlue',
                 'Purple', 'Grey', 'Yellow', 'LightGreen', 'DeepOrange',
                 'Green', 'Red', 'Teal', 'Orange', 'Cyan', 'Amber',
                 'DeepPurple', 'Lime'])

    primary_hue = OptionProperty(
        '500',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    primary_light_hue = OptionProperty(
        '200',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    primary_dark_hue = OptionProperty(
        '700',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    def _get_primary_color(self):
        return get_color_from_hex(
            colors[self.primary_palette][self.primary_hue])

    primary_color = AliasProperty(_get_primary_color,
                                  bind=('primary_palette', 'primary_hue'))

    def _get_primary_light(self):
        return get_color_from_hex(
            colors[self.primary_palette][self.primary_light_hue])

    primary_light = AliasProperty(
        _get_primary_light, bind=('primary_palette', 'primary_light_hue'))

    def _get_primary_dark(self):
        return get_color_from_hex(
            colors[self.primary_palette][self.primary_dark_hue])

    primary_dark = AliasProperty(_get_primary_dark,
                                 bind=('primary_palette', 'primary_dark_hue'))

    accent_palette = OptionProperty(
        'Amber',
        options=['Pink', 'Blue', 'Indigo', 'BlueGrey', 'Brown',
                 'LightBlue',
                 'Purple', 'Grey', 'Yellow', 'LightGreen', 'DeepOrange',
                 'Green', 'Red', 'Teal', 'Orange', 'Cyan', 'Amber',
                 'DeepPurple', 'Lime'])

    accent_hue = OptionProperty(
        '500',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    accent_light_hue = OptionProperty(
        '200',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    accent_dark_hue = OptionProperty(
        '700',
        options=['50', '100', '200', '300', '400', '500', '600', '700',
                 '800',
                 '900', 'A100', 'A200', 'A400', 'A700'])

    def _get_accent_color(self):
        return get_color_from_hex(
            colors[self.accent_palette][self.accent_hue])

    accent_color = AliasProperty(_get_accent_color,
                                 bind=['accent_palette', 'accent_hue'])

    def _get_accent_light(self):
        return get_color_from_hex(
            colors[self.accent_palette][self.accent_light_hue])

    accent_light = AliasProperty(_get_accent_light,
                                 bind=['accent_palette', 'accent_light_hue'])

    def _get_accent_dark(self):
        return get_color_from_hex(
            colors[self.accent_palette][self.accent_dark_hue])

    accent_dark = AliasProperty(_get_accent_dark,
                                bind=['accent_palette', 'accent_dark_hue'])

    theme_style = OptionProperty('Light', options=['Light', 'Dark'])

    def _get_theme_style(self, opposite):
        if opposite:
            return 'Light' if self.theme_style == 'Dark' else 'Dark'
        else:
            return self.theme_style

    def _get_bg_darkest(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            return get_color_from_hex(colors['Light']['StatusBar'])
        elif theme_style == 'Dark':
            return get_color_from_hex(colors['Dark']['StatusBar'])

    bg_darkest = AliasProperty(_get_bg_darkest, bind=['theme_style'])

    def _get_op_bg_darkest(self):
        return self._get_bg_darkest(True)

    opposite_bg_darkest = AliasProperty(_get_op_bg_darkest,
                                        bind=['theme_style'])

    def _get_bg_dark(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            return get_color_from_hex(colors['Light']['AppBar'])
        elif theme_style == 'Dark':
            return get_color_from_hex(colors['Dark']['AppBar'])

    bg_dark = AliasProperty(_get_bg_dark, bind=['theme_style'])

    def _get_op_bg_dark(self):
        return self._get_bg_dark(True)

    opposite_bg_dark = AliasProperty(_get_op_bg_dark, bind=['theme_style'])

    def _get_bg_normal(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            return get_color_from_hex(colors['Light']['Background'])
        elif theme_style == 'Dark':
            return get_color_from_hex(colors['Dark']['Background'])

    bg_normal = AliasProperty(_get_bg_normal, bind=['theme_style'])

    def _get_op_bg_normal(self):
        return self._get_bg_normal(True)

    opposite_bg_normal = AliasProperty(_get_op_bg_normal, bind=['theme_style'])

    def _get_bg_light(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            return get_color_from_hex(colors['Light']['CardsDialogs'])
        elif theme_style == 'Dark':
            return get_color_from_hex(colors['Dark']['CardsDialogs'])

    bg_light = AliasProperty(_get_bg_light, bind=['theme_style'])

    def _get_op_bg_light(self):
        return self._get_bg_light(True)

    opposite_bg_light = AliasProperty(_get_op_bg_light, bind=['theme_style'])

    def _get_divider_color(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            color = get_color_from_hex('000000')
        elif theme_style == 'Dark':
            color = get_color_from_hex('FFFFFF')
        color[3] = .12
        return color

    divider_color = AliasProperty(_get_divider_color, bind=['theme_style'])

    def _get_op_divider_color(self):
        return self._get_divider_color(True)

    opposite_divider_color = AliasProperty(_get_op_divider_color,
                                           bind=['theme_style'])

    def _get_text_color(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            color = get_color_from_hex('000000')
            color[3] = .87
        elif theme_style == 'Dark':
            color = get_color_from_hex('FFFFFF')
        return color

    text_color = AliasProperty(_get_text_color, bind=['theme_style'])

    def _get_op_text_color(self):
        return self._get_text_color(True)

    opposite_text_color = AliasProperty(_get_op_text_color,
                                        bind=['theme_style'])

    def _get_secondary_text_color(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            color = get_color_from_hex('000000')
            color[3] = .54
        elif theme_style == 'Dark':
            color = get_color_from_hex('FFFFFF')
            color[3] = .70
        return color

    secondary_text_color = AliasProperty(_get_secondary_text_color,
                                         bind=['theme_style'])

    def _get_op_secondary_text_color(self):
        return self._get_secondary_text_color(True)

    opposite_secondary_text_color = AliasProperty(_get_op_secondary_text_color,
                                                  bind=['theme_style'])

    def _get_icon_color(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            color = get_color_from_hex('000000')
            color[3] = .54
        elif theme_style == 'Dark':
            color = get_color_from_hex('FFFFFF')
        return color

    icon_color = AliasProperty(_get_icon_color,
                               bind=['theme_style'])

    def _get_op_icon_color(self):
        return self._get_icon_color(True)

    opposite_icon_color = AliasProperty(_get_op_icon_color,
                                        bind=['theme_style'])

    def _get_disabled_hint_text_color(self, opposite=False):
        theme_style = self._get_theme_style(opposite)
        if theme_style == 'Light':
            color = get_color_from_hex('000000')
            color[3] = .38
        elif theme_style == 'Dark':
            color = get_color_from_hex('FFFFFF')
            color[3] = .50
        return color

    disabled_hint_text_color = AliasProperty(_get_disabled_hint_text_color,
                                             bind=['theme_style'])

    def _get_op_disabled_hint_text_color(self):
        return self._get_disabled_hint_text_color(True)

    opposite_disabled_hint_text_color = AliasProperty(
        _get_op_disabled_hint_text_color, bind=['theme_style'])

    # Hardcoded because muh standard
    def _get_error_color(self):
        return get_color_from_hex(colors['Red']['A700'])

    error_color = AliasProperty(_get_error_color)


    def _get_ripple_color(self):
        return self._ripple_color

    def _set_ripple_color(self, value):
        self._ripple_color = value

    _ripple_color = ListProperty(get_color_from_hex(colors['Grey']['400']))
    ripple_color = AliasProperty(_get_ripple_color,
                                 _set_ripple_color,
                                 bind=['_ripple_color'])

    def _determine_device_orientation(self, _, window_size):
        if window_size[0] > window_size[1]:
            self.device_orientation = 'landscape'
        elif window_size[1] >= window_size[0]:
            self.device_orientation = 'portrait'

    device_orientation = StringProperty('')

    def _get_standard_increment(self):
        if DEVICE_TYPE == 'mobile':
            if self.device_orientation == 'landscape':
                return dp(48)
            else:
                return dp(56)
        else:
            return dp(64)

    standard_increment = AliasProperty(_get_standard_increment,
                                       bind=['device_orientation'])

    def _get_horizontal_margins(self):
        if DEVICE_TYPE == 'mobile':
            return dp(16)
        else:
            return dp(24)

    horizontal_margins = AliasProperty(_get_horizontal_margins)

    def on_theme_style(self, instance, value):
        if hasattr(App.get_running_app(), 'theme_cls') and \
                        App.get_running_app().theme_cls == self:
            self.set_clearcolor_by_theme_style(value)

    def set_clearcolor_by_theme_style(self, theme_style):
        if theme_style == 'Light':
            Window.clearcolor = get_color_from_hex(
                colors['Light']['Background'])
        elif theme_style == 'Dark':
            Window.clearcolor = get_color_from_hex(
                colors['Dark']['Background'])

    def __init__(self, **kwargs):
        super(ThemeManager, self).__init__(**kwargs)
        self.rec_shadow = Atlas('{}rec_shadow.atlas'.format(images_path))
        self.rec_st_shadow = Atlas('{}rec_st_shadow.atlas'.format(images_path))
        self.quad_shadow = Atlas('{}quad_shadow.atlas'.format(images_path))
        self.round_shadow = Atlas('{}round_shadow.atlas'.format(images_path))
        Clock.schedule_once(lambda x: self.on_theme_style(0, self.theme_style))
        self._determine_device_orientation(None, Window.size)
        Window.bind(size=self._determine_device_orientation)


class ThemableBehavior(object):
    theme_cls = ObjectProperty(None)
    opposite_colors = BooleanProperty(False)

    def __init__(self, **kwargs):
        if self.theme_cls is not None:
            pass
        elif hasattr(App.get_running_app(), 'theme_cls'):
            self.theme_cls = App.get_running_app().theme_cls
        else:
            self.theme_cls = ThemeManager()
        super(ThemableBehavior, self).__init__(**kwargs)



from traceback import print_exc
from kivy.logger import Logger
from kivy.factory import Factory
from kivy.app import App
from kivy.core.window import Window
from kivy.uix.popup import Popup
from appPublic.Singleton import SingletonDecorator
from appPublic.jsonConfig import getConfig
from appPublic.registerfunction import RegisterFunction
logformdesc = {
	"widgettype":"Form",
	"options":{
		"cols":1,
		"labelwidth":0.4,
		"textsize":1,
		"inputheight":4,
		"fields":[
			{
				"name":"userid",
				"label":"user name",
				"datatype":"str",
				"required":True,
				"uitype":"string"
			},
			{
				"name":"passwd",
				"label":"Password",
				"datatype":"str",
				"uitype":"password"
			}
		]
	}
}

@SingletonDecorator
class LoginForm(Popup):
	def __init__(self):
		super().__init__(size_hint=(0.8,0.8))
		self.title = 'login'
		self.buildContent(None,None)
		self.register_event_type('on_setupuser')
	
	def on_setupuser(self,o=None):
		return

	def buildContent(self,o,size):
		self.content = Factory.Form(**logformdesc['options'])
		self.content.bind(on_submit=self.on_submit)
		
	def on_submit(self,o,userinfo):
		print('login(),on_submit fired ....')
		self.dismiss()
		print('userinfo=',userinfo)
		app = App.get_running_app()

		if userinfo.get('passwd',False):
			userinfo['authmethod'] = 'password'
		authinfo = app.serverinfo.encode(userinfo)
		self.dispatch('on_setupuser')
	
	def on_cancel(self,o,v):
		print('login() on_cancel fired .....')
		self.dismiss()






from ..baseWidget import Password
from ..utils import *
from .factory import UiFactory, get_value, StringView

class PasswordView(StringView):
	def __init__(self, text='*', **kw):
		super(PasswordView, self).__init__(text=text, **kw)

	def getValue(self):
		return None

	def setValue(self, v):
		pass

	def builder(self, desc, rec=None):
		return PasswordView(text='*',
						font_size=CSize(1),
						halign='left',
						valign='middle')

def build_input_password_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = Password(**kw)
	if v:
		w.setValue(v)
	return w
	
UiFactory.register('password', build_input_password_widget, PasswordView.builder)




from kivy.properties import NumericProperty, StringProperty, \
			OptionProperty, ListProperty
from kivy.uix.scrollview import ScrollView
from kivy.effects.scroll import ScrollEffect
from kivy.uix.widget import Widget
from kivy.uix.boxlayout import BoxLayout
from kivy.factory import Factory

from .utils import *
from .widget_css import WidgetCSS

class ScrollPanel(WidgetCSS, ScrollView):
	orientation = OptionProperty('vertical', \
						options=['vertical', 'horizontal'])
	bgcolor = ListProperty([0.2, 0.2, 0.2, 1])
	padding_c = NumericProperty(1)
	def __init__(self,inner=None, **kw):
		SUPER(ScrollPanel, self, kw)
		self.effect_cls = ScrollEffect
		if not inner:
			self.create_default_inner()
		elif isinstance(inner, Widget):
			self._inner = inner
		else:
			self._inner = Factory.Blocks().widgetBuild(inner)

		super(ScrollPanel,self).add_widget(self._inner)
		self._inner.bind(
				minimum_height=self._inner.setter('height'))
		self._inner.bind(
				minimum_width=self._inner.setter('width'))

	def get_subwidgets(self):
		return self._inner.children

	def create_default_inner(self):
		kw = {
			'size_hint':(None,None),
			'orientation':self.orientation
		}
		desc = {
			"widgettype":"Box",
			"options":kw
		}
		self._inner = Factory.Blocks().widgetBuild(desc)
		self._inner.bgcolor = self.bgcolor
		if not self._inner:
			print('desc=', desc)
			raise Exception('widget build failed')
		self._inner.spacing = CSize(self.padding_c)

	def add_widget(self,widget,**kw):
		a = self._inner.add_widget(widget,**kw)
		return a

	def clear_widgets(self,**kw):
		a = self._inner.clear_widgets(**kw)

	def remove_widget(self,widget,**kw):
		a = self._inner.remove_widget(widget,**kw)
		return a

if __name__ == '__main__':
	from kivy.app import App
	from kivy.uix.label import Label
	from kivy.uix.button import Button
	import codecs
	
	class MyApp(App):
		def build(self):
			root = ScrollPanel(size=(400,400),
					pos_hint={'center_x': .5, 'center_y': .5}
			)
			with codecs.open(__file__,'r','utf-8') as f:
				txt = f.read()
				lines = txt.split('\n')
				for l in lines:
					root.add_widget(Label(text=l,color=(1,1,1,1),size_hint=(None,None),size=(1200,40)))
			return root

	MyApp().run()



from kivy.logger import Logger
from kivy.graphics import Color, Rectangle, RoundedRectangle
from kivy.properties import ListProperty, NumericProperty
from .color_definitions import getColors

_logcnt = 0
class BGColorBehavior(object):
	color_level=NumericProperty(0)
	radius = ListProperty([])
	def __init__(self, **kwargs):
		self.bg_ready = False
		self.bgcolor = []
		self.fgcolor = []
		fg,bg= getColors(self.color_level)
		self.fgcolor = fg
		self.bgcolor = bg
		self.normal_bgcolor = bg
		self.normal_fgcolor = fg
		fg,bg= getColors(self.color_level,selected=True)
		self.selected_bgcolor = bg
		self.selected_fgcolor = fg
		# self.on_bgcolor()

		self.bind(size=self.onSize_bgcolor_behavior,
				pos=self.onSize_bgcolor_behavior)
		# self.bind(children=self.on_bgcolor)

	def onSize_bgcolor_behavior(self,o,v=None):
		if not hasattr(self,'rect'):
			self.on_bgcolor()
		else:
			self.rect.pos = self.pos
			self.rect.size = self.size

	def on_bgcolor(self,o=None,v=None):
		if self.bgcolor == []:
			return

		if self.canvas:
			with self.canvas.before:
				Color(*self.bgcolor)
				if self.radius != []:
					self.rect = RoundedRectangle(pos=self.pos,
								size=self.size,
								radius=self.radius)
				else:
					self.rect = Rectangle(pos=self.pos, 
								size=self.size)
		else:
			print('on_bgcolor():self.canvas is None')

	def is_selected(self):
		return self.bgcolor == self.selected_bgcolor

	def selected(self):
		if self.bgcolor == self.selected_bgcolor:
			return
		self.bgcolor = self.selected_bgcolor
		self.color = self.fgcolor = self.selected_fgcolor
		self.on_bgcolor()

	def unselected(self):
		if self.bgcolor == self.normal_bgcolor:
			return
		self.bgcolor = self.normal_bgcolor
		self.color = self.fgcolor = self.normal_fgcolor
		self.on_bgcolor()



import os
from traceback import print_exc
from traceback import print_exc
from kivy.app import App
from kivy.logger import Logger
from appPublic.jsonConfig import getConfig
from kivy.uix.popup import Popup
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.modalview import ModalView
from kivy.uix.image import AsyncImage
from appPublic.dictObject import DictObject

from .kivysize import KivySizes

class NeedLogin(Exception):
	pass

class InsufficientPrivilege(Exception):
	pass

class HTTPError(Exception):
	def __init__(self,resp_code):
		self.resp_code = resp_code
		Exception.__init__(self)

	def __expr__(self):
		return f'Exception:return code={self.resp_code}'

	def __str__(self):
		return f'Exception:return code={self.resp_code}'

alert_widget= None

def kwarg_pop(obj, kw):
	keys = [k for k in kw.keys()]
	for k in keys:
		if hasattr(obj, k):
			setattr(obj, k, kw.pop(k))

def SUPER(klass, obj, kw):
	keys = [ k for k in kw.keys() ]
	dic = { k:kw.pop(k) for k in keys if hasattr(obj, k) }
	super(klass, obj).__init__(**kw)
	for k,v in dic.items():
		try:
			setattr(obj, k, v)
		except Exception as e:
			print(f'obj={obj}, setattr(obj, "{k}","{v}") error')
			print_exc()
			raise e

def blockImage(name):
	p = os.path.dirname(os.path.abspath(__file__))
	return os.path.join(p,'imgs',name)

def loaded(widget):
	widget.loadingwidget.dismiss()
	# widget.remove_widget(widget.loadingwidget)
	del widget.loadingwidget
	widget.loadingwidget = None

def loading(parent):
	fp = os.path.join(os.path.dirname(__file__),'imgs','loading1.gif')
	image = AsyncImage(source=blockImage('loading1.gif'), \
					width=CSize(2), height=CSize(2),
					size_hint=(None,None))
	view = ModalView(auto_dismiss=False)
	view.add_widget(image)
	view.center = parent.center
	parent.loadingwidget = view
	# parent.add_widget(view)
	view.open()
	return view

def set_widget_width(self, width):
	if width <= 1:
		self.size_hint_x = width
	else:
		self.size_hint_x = None
		self.width = width

def set_widget_height(self, height):
	if height <= 1:
		self.size_hint_y = height
	else:
		self.size_hint_y = None
		self.height = height

def setSizeOptions(desc,kw):
	"""
	desc's width, and height to setup a widget's size options 
	if width or height is not set, kw add not thing 
	if width or height <= 1, using present rate of size
	else use CSize to tims width or height
	"""
	if not isinstance(kw, DictObject):
		kw = DictObject(**kw)
	
	width = desc.get('width',0)
	if width > 1.01:
		kw.width = CSize(width)
		kw.size_hint_x = None
	elif width > 0.00:
		kw.size_hint_x = width

	height = desc.get('height',0)
	if height > 1.01:
		kw.height = CSize(height)
		kw.size_hint_y = None
	elif height > 0.00:
		kw.size_hint_y = height
	return kw
	
def alert(text,title='alert'):
	global alert_widget
	def close_alert(obj):
		alert_widget.dismiss()

	charsize = CSize(1)
	if alert_widget is None:
		bl = BoxLayout(orientation='horizontal')
		msg = Label(font_size=charsize)
		bl.add_widget(msg)
		button = Button(size_hint_y=None,height=1.4*charsize,font_size=charsize,text='OK')
		button.bind(on_press=close_alert)
		bl.add_widget(button)
		alert_widget = Popup(content=bl, size_hint=(0.9,0.6))
		alert_widget.msg_widget = msg
	alert_widget.msg_widget.text = str(text)
	alert_widget.title = str(title)
	alert_widget.open()

def StrConvert(s):
	if not s.startswith('py::'):
		return s
	s = s[4:]
	try:
		ns = {}
		exec('_n_=' + s,globals(),ns)
		return ns['_n_']
	except Exception as e:
		print('----e=',e,'------------s=',s)
		return s

def ArrayConvert(a):
	s = []
	for i in a:
		s.append(JDConvert(i))
	return s

def DictConvert(dic):
	d = {}
	for k,v in dic.items():
		if k == 'widgettype':
			d[k] = v
		else:
			d[k] = JDConvert(v)
	return d

def JDConvert(dic):
	nd = {}
	if type(dic) == type(''):
		return StrConvert(dic)
	if type(dic) == type([]):
		return ArrayConvert(dic)
	if type(dic) == type({}):
		return DictConvert(dic)
	return dic
				
def getWidgetById(w,id):
	if id[0] == '/':
		app = App.get_running_ap()
		if not hasattr('ids'):
			return None
		return app.ids.get(id[1:])
	if id in ['self', '.' ]:
		return w
	if not hasattr(w,'ids'):
		return None
	return w.ids.get(id)
		
def CSize(x,y=None,name=None):
	ks = KivySizes()
	return ks.CSize(x,y=y,name=name)

def screenSize():
	ks = KivySizes()
	return ks.getScreenSize()

def screenPhysicalSize():
	ks = KivySizes()
	return ks.getScreenPhysicalSize()

def isHandHold():
	ks = KivySizes()
	return ks.isHandHold()

def absurl(url,parent):
	if parent is None:
		parent = ''
	config = getConfig()
	if url.startswith('http://'):
		return url
	if url.startswith('https://'):
		return url
	if url.startswith('file:///'):
		return url
	if url.startswith('/'):
		return config.uihome + url
	if url.startswith(config.uihome):
		return url
	if parent == '':
		print('url=',url)
		raise Exception('related url(%s) need a parent url' % url)

	if parent.startswith(config.uihome):
		parent = parent[len(config.uihome):]
	paths = parent.split('/')
	paths.pop()
	for i in url.split('/'):
		if i in [ '.', '' ]:
			continue
		if i == '..':
			if len(paths) > 1:
				paths.pop()
			continue
		paths.append(i)
	return config.uihome + '/'.join(paths)

def show_widget_info(w, tag='DEBUG'):
	id = getattr(w, 'widget_id', 'null')
	msg=f"""{tag}:size_hint={w.size_hint},size={w.size},pos={w.pos},widget_id={id},{w}"""
	Logger.info(msg)



from kivy.uix.floatlayout import FloatLayout
from kivy.event import EventDispatcher
from kivy.clock import Clock
from kivy.factory import Factory
from kivy.properties import ObjectProperty
from kivy.uix.popup import Popup
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.filechooser import FileChooserListView, FileChooserIconView
from kivyblocks.i18n import I18n
from kivyblocks.utils import blockImage,CSize
from kivyblocks.baseWidget import PressableImage, StrInput

import os

class FileLoaderBrowser(Popup):
	def __init__(self,rootpath='/', title='load file', choose_dir=False, **kwargs):
		i18n = I18n()
		self.content = BoxLayout(orientation='vertical')
		Popup.__init__(self, content = self.content, 
					title=i18n(title), 
					size_hint=(0.9,0.9))
		self.register_event_type('on_selected')
		self.choose_dir = choose_dir

		self.list_widget = FileChooserListView(rootpath=rootpath)
		self.icon_widget = FileChooserIconView(rootpath=rootpath)
		self.list_widget.bind(on_touch_up=self.on_release)
		self.icon_widget.bind(on_touch_up=self.on_release)
		self.list_widget.bind(on_submit=self.do_submit)
		self.icon_widget.bind(on_submit=self.do_submit)
		self.btn_load = Button(text=i18n('load'),size_hint=(None,None),
						size=CSize(8,2))
		self.btn_list_view = PressableImage(	\
						source=blockImage('list_view.png'),
						size_hint=(None,None),
						size=CSize(2,2))
		self.btn_icon_view = PressableImage(	\
						source=blockImage('icon_view.png'),
						size_hint=(None,None),
						size=CSize(2,2))
		self.btn_list_view.bind(on_press=self.change_view_mode)
		self.btn_icon_view.bind(on_press=self.change_view_mode)
		self.btn_load.bind(on_release=self.do_load)
		self.control_box = BoxLayout(orientation='horizontal',
				size_hint_y=None,
				height=CSize(2.5)
				)
		self.txt_input = Factory.StrInput()
		self.view_mode = 'list_view'
		self.show_list_view()

	def on_selected(self,fn):
		print(fn,'selected')

	def on_release(self,o,v=None):
		Clock.schedule_once(self.set_txt_input,0.4)

	def set_txt_input(self,t=None):
		self.txt_input.text = self.cur_widget.path

	def target_selected(self):
		fn = self.txt_input.text
		if not self.choose_dir and os.path.isdir(fn):
			return 
		if self.choose_dir and os.path.isfile(fn):
			fn = os.path.dirname(fn)
		self.dispatch('on_selected',fn)
		self.dismiss()

	def do_submit(self,o,a,b):
		if not self.cur_widget.selection:
			return
		self.txt_input.text = self.cur_widget.selection[0]
		self.target_selected()

	def do_load(self,o,d=None):
		if self.cur_widget.selection:
			self.txt_input.text = self.cur_widget.selection[0]
		if self.txt_input.text == '':
			return
		self.target_selected()

	def change_view_mode(self,o,v=None):
		self.content.clear_widgets()
		if self.view_mode == 'list_view':
			self.show_icon_view()
		else:
			self.show_list_view()

	def show_list_view(self):
		self.cur_widget = self.list_widget
		self.view_mode = 'list_view'
		self.cur_widget.path = self.icon_widget.path
		self.content.add_widget(self.list_widget)
		self.content.add_widget(self.txt_input)
		self.content.add_widget(self.control_box)
		self.control_box.clear_widgets()
		self.control_box.add_widget(self.btn_load)
		self.control_box.add_widget(self.btn_list_view)

	def show_icon_view(self):
		self.cur_widget = self.icon_widget
		self.view_mode = 'icon_view'
		self.cur_widget.path = self.list_widget.path
		self.content.add_widget(self.icon_widget)
		self.content.add_widget(self.txt_input)
		self.content.add_widget(self.control_box)
		self.control_box.clear_widgets()
		self.control_box.add_widget(self.btn_load)
		self.control_box.add_widget(self.btn_icon_view)

if __name__ == '__main__':
	from kivy.app import App
	class Test(App):
		def build(self):
			x = FileLoaderBrowser()
			return x
	app = Test()
	app.run()
